import logging

import pytest
import pytest_asyncio

from gaphor import UML
from gaphor.application import Application
from gaphor.core import Transaction, event_handler
from gaphor.core.modeling import AssociationUpdated, Diagram
from gaphor.diagram.copypaste import copy_full, paste_full
from gaphor.diagram.tests.fixtures import connect
from gaphor.UML.classes import (
    AssociationItem,
    ClassItem,
    ContainmentItem,
    GeneralizationItem,
)
from gaphor.UML.interactions import MessageItem
from gaphor.UML.interactions.interactionstoolbox import reflexive_message_config


@pytest.fixture
def application():
    app = Application()
    yield app
    app.shutdown()


@pytest_asyncio.fixture
async def session(application):
    return application.new_session()


@pytest.fixture
def event_manager(session):
    return session.get_service("event_manager")


@pytest.fixture
def element_factory(session):
    return session.get_service("element_factory")


@pytest.fixture
def undo_manager(session):
    return session.get_service("undo_manager")


def test_class_association_undo_redo(event_manager, element_factory, undo_manager):
    diagram, ci1, ci2, a = set_up_class_and_association(event_manager, element_factory)

    def get_connected(handle):
        """Get item connected to line via handle."""
        if cinfo := diagram.connections.get_connection(handle):
            return cinfo.connected
        return None

    undo_manager.clear_undo_stack()
    assert not undo_manager.can_undo()

    with Transaction(event_manager):
        ci2.unlink()

    assert undo_manager.can_undo()

    for _i in range(3):
        assert 9 == len(diagram.connections.solver.constraints)

        undo_manager.undo_transaction()

        assert 18 == len(diagram.connections.solver.constraints)

        a = next(element_factory.select(AssociationItem))
        assert ci1 == get_connected(a.head)
        assert ci2.id == get_connected(a.tail).id

        undo_manager.redo_transaction()


def set_up_class_and_association(event_manager, element_factory):
    with Transaction(event_manager):
        diagram = element_factory.create(UML.Diagram)

    assert 0 == len(diagram.connections.solver.constraints)

    with Transaction(event_manager):
        ci1 = diagram.create(ClassItem, subject=element_factory.create(UML.Class))
    assert 8 == len(diagram.connections.solver.constraints)

    with Transaction(event_manager):
        ci2 = diagram.create(ClassItem, subject=element_factory.create(UML.Class))
    assert 16 == len(diagram.connections.solver.constraints)

    with Transaction(event_manager):
        a = diagram.create(AssociationItem)

        connect(a, a.head, ci1)
        connect(a, a.tail, ci2)

    # Diagram, Association, 2x Class, Property, LiteralSpecification
    assert 9 == len(element_factory.lselect())
    assert 18 == len(diagram.connections.solver.constraints)

    return diagram, ci1, ci2, a


def test_diagram_item_can_undo_and_redo(
    event_manager, element_factory, undo_manager, caplog
):
    caplog.set_level(logging.INFO)
    with Transaction(event_manager):
        diagram = element_factory.create(Diagram)

    with Transaction(event_manager):
        cls = diagram.create(ClassItem, subject=element_factory.create(UML.Class))
        cls.subject.name = "name"
        cls.matrix.translate(10, 10)

    undo_manager.undo_transaction()
    undo_manager.redo_transaction()
    new_cls = diagram.ownedPresentation[0]
    assert new_cls.matrix.tuple() == (1, 0, 0, 1, 10, 10)
    assert new_cls.subject, element_factory.select()
    assert new_cls.subject.name == "name"
    assert not caplog.records


def test_diagram_item_should_not_end_up_in_element_factory(
    event_manager, element_factory, undo_manager
):
    with Transaction(event_manager):
        diagram = element_factory.create(Diagram)

    with Transaction(event_manager):
        cls = diagram.create(ClassItem, subject=element_factory.create(UML.Class))

    undo_manager.undo_transaction()
    undo_manager.redo_transaction()

    assert cls not in element_factory.lselect(), element_factory.lselect()


def test_delete_and_undo_diagram_item(event_manager, element_factory, undo_manager):
    with Transaction(event_manager):
        diagram = element_factory.create(Diagram)

    with Transaction(event_manager):
        subject = element_factory.create(UML.Class)
        subject.name = "Name"
        cls = diagram.create(ClassItem, subject=subject)

    with Transaction(event_manager):
        cls.unlink()

    undo_manager.undo_transaction()

    new_cls = diagram.ownedPresentation[0]
    new_elem = element_factory.lookup(subject.id)

    assert new_cls in new_elem.presentation
    assert new_cls.subject
    assert new_elem.name == "Name"


def test_delete_and_undo_model_element(event_manager, element_factory, undo_manager):
    with Transaction(event_manager):
        diagram = element_factory.create(Diagram)

    with Transaction(event_manager):
        subject = element_factory.create(UML.Class)
        subject.name = "Name"
        diagram.create(ClassItem, subject=subject)

    with Transaction(event_manager):
        subject.unlink()

    undo_manager.undo_transaction()

    new_cls = diagram.ownedPresentation[0]
    new_elem = element_factory.lookup(subject.id)

    assert new_cls in new_elem.presentation
    assert new_cls.subject
    assert new_elem.name == "Name"


def test_deleted_diagram_item_should_not_end_up_in_element_factory(
    event_manager, element_factory, undo_manager
):
    with Transaction(event_manager):
        diagram = element_factory.create(Diagram)
        cls = diagram.create(ClassItem, subject=element_factory.create(UML.Class))

    with Transaction(event_manager):
        cls.unlink()

    undo_manager.undo_transaction()

    assert cls not in element_factory.lselect(), element_factory.lselect()

    undo_manager.redo_transaction()

    assert cls not in element_factory.lselect(), element_factory.lselect()


def test_undo_should_not_cause_warnings(
    event_manager, element_factory, undo_manager, caplog
):
    caplog.set_level(logging.INFO)
    with Transaction(event_manager):
        diagram = element_factory.create(Diagram)

    with Transaction(event_manager):
        diagram.create(ClassItem, subject=element_factory.create(UML.Class))

    assert not caplog.records

    undo_manager.undo_transaction()

    assert not diagram.ownedPresentation
    assert not caplog.records


def test_can_undo_connected_generalization(
    event_manager, element_factory, undo_manager, caplog
):
    caplog.set_level(logging.INFO)
    with Transaction(event_manager):
        diagram: Diagram = element_factory.create(Diagram)
        general = diagram.create(ClassItem, subject=element_factory.create(UML.Class))
        specific = diagram.create(ClassItem, subject=element_factory.create(UML.Class))

    with Transaction(event_manager):
        generalization = diagram.create(GeneralizationItem)
        connect(generalization, generalization.head, general)
        connect(generalization, generalization.tail, specific)

    assert not caplog.records

    undo_manager.undo_transaction()

    assert not list(diagram.select(GeneralizationItem))
    assert not caplog.records

    undo_manager.redo_transaction()
    new_generalization_item = next(diagram.select(GeneralizationItem))
    new_generalization = next(element_factory.select(UML.Generalization))

    assert len(list(diagram.select(GeneralizationItem))) == 1
    assert len(element_factory.lselect(UML.Generalization)) == 1
    assert new_generalization_item.subject is new_generalization
    assert not caplog.records


def test_can_undo_connected_association(
    event_manager, element_factory, undo_manager, caplog
):
    caplog.set_level(logging.INFO)
    with Transaction(event_manager):
        diagram: UML.Diagram = element_factory.create(UML.Diagram)
        parent = diagram.create(ClassItem, subject=element_factory.create(UML.Class))
        child = diagram.create(ClassItem, subject=element_factory.create(UML.Class))

    with Transaction(event_manager):
        association = diagram.create(AssociationItem)
        connect(association, association.head, parent)
        connect(association, association.tail, child)

    assert not caplog.records

    undo_manager.undo_transaction()

    assert not list(diagram.select(AssociationItem))
    assert not caplog.records

    undo_manager.redo_transaction()
    new_association_item = next(diagram.select(AssociationItem))
    new_association = next(element_factory.select(UML.Association))

    assert len(list(diagram.select(AssociationItem))) == 1
    assert len(element_factory.lselect(UML.Association)) == 1

    assert len(new_association.memberEnd) == 2
    assert new_association_item.subject is new_association
    assert new_association_item.head_subject
    assert new_association_item.tail_subject
    assert not caplog.records


def test_can_undo_diagram_with_content(event_manager, element_factory, undo_manager):
    with Transaction(event_manager):
        diagram: Diagram = element_factory.create(Diagram)
        diagram.create(ClassItem, subject=element_factory.create(UML.Class))

    with Transaction(event_manager):
        diagram.unlink()

    undo_manager.undo_transaction()

    new_diagram = element_factory.lookup(diagram.id)

    assert new_diagram
    assert new_diagram.ownedPresentation


def test_reflexive_message_undo(event_manager, element_factory, undo_manager):
    with Transaction(event_manager):
        diagram: Diagram = element_factory.create(Diagram)

    with Transaction(event_manager):
        message = diagram.create(MessageItem)
        reflexive_message_config(message)

    undo_manager.undo_transaction()


def test_delete_item_with_subject_owning_diagram(
    event_manager, element_factory, undo_manager
):
    with Transaction(event_manager):
        diagram: UML.Diagram = element_factory.create(UML.Diagram)
        klass = element_factory.create(UML.Class)
        class_item = diagram.create(ClassItem, subject=klass)
        diagram.element = klass

    with Transaction(event_manager):
        class_item.unlink()

    undo_manager.undo_transaction()


def test_reconnect_on_same_element(event_manager, element_factory, undo_manager):
    def copy_pos(pos):
        return tuple(map(float, pos))

    with Transaction(event_manager):
        diagram: Diagram = element_factory.create(Diagram)
        klass = element_factory.create(UML.Class)
        class_item = diagram.create(ClassItem, subject=klass)
        association = diagram.create(AssociationItem)
        connect(association, association.head, class_item)
        original_handle_pos = copy_pos(association.head.pos)

    with Transaction(event_manager):
        association.head.pos = (class_item.width, class_item.height)
        connect(association, association.head, class_item)
        new_handle_pos = copy_pos(association.head.pos)

    assert original_handle_pos != new_handle_pos

    undo_manager.undo_transaction()
    diagram.update(diagram.ownedPresentation)

    assert original_handle_pos == copy_pos(association.head.pos)


def test_exception_raised_during_undo(event_manager, element_factory, undo_manager):
    with Transaction(event_manager):
        package = element_factory.create(UML.Package)
        klass = element_factory.create(UML.Class)
        klass.package = package

    with Transaction(event_manager):
        klass.unlink()

    @event_handler(AssociationUpdated)
    def raise_an_exception(event):
        if event.property is UML.Class.package:
            raise ValueError("Test exception")

    event_manager.subscribe(raise_an_exception)

    with pytest.raises(ExceptionGroup):
        undo_manager.undo_transaction()


def test_exception_raised_during_undo_from_event_handler(
    event_manager, element_factory, undo_manager
):
    class UndoRequested:
        pass

    with Transaction(event_manager):
        package = element_factory.create(UML.Package)
        klass = element_factory.create(UML.Class)
        klass.package = package

    with Transaction(event_manager):
        klass.unlink()

    @event_handler(AssociationUpdated)
    def raise_an_exception(event):
        if event.property is UML.Class.package:
            raise ValueError("Test exception")

    event_manager.subscribe(raise_an_exception)

    @event_handler(UndoRequested)
    def on_undo_requested(event):
        undo_manager.undo_transaction()

    event_manager.subscribe(on_undo_requested)

    with pytest.raises(ExceptionGroup):
        event_manager.handle(UndoRequested())

    # Handle remaining events on the queue
    event_manager.handle()


def test_undo_paste_full(create, diagram, event_manager, undo_manager):
    """Triggers an error when pasted data is undone.

    Reproduce Hypothesis test found in https://github.com/gaphor/gaphor/issues/2895.
    """

    with Transaction(event_manager):
        class_item = create(ClassItem, UML.Class)
        containment_item = create(ContainmentItem)

        connect(containment_item, containment_item.head, class_item)

    copy_buffer = copy_full([containment_item, class_item])

    with Transaction(event_manager):
        paste_full(copy_buffer, diagram)

    undo_manager.undo_transaction()



from pathlib import Path

import pytest

import gaphor.services.properties
from gaphor.event import ModelSaved, SessionCreated


@pytest.fixture
def properties(event_manager):
    properties = gaphor.services.properties.Properties(event_manager)
    yield properties
    properties.shutdown()


def test_set_property(properties):
    properties.on_model_loaded(SessionCreated(None, None, filename="some_file_name"))

    properties.set("test", 1)

    assert properties.get("test") == 1


def test_load_properties(properties, event_manager):
    properties.set("test", 1)
    properties.on_model_saved(ModelSaved("test_load_properties"))

    new_properties = gaphor.services.properties.Properties(event_manager)
    new_properties.on_model_loaded(SessionCreated(None, None, "test_load_properties"))

    assert new_properties.get("test") == 1


def test_load_of_corrupted_properties(properties, event_manager, caplog):
    properties.set("test", 1)
    properties.on_model_saved(ModelSaved("test_load_properties"))
    Path(properties.filename).write_text("{ invalid content }", encoding="utf-8")

    new_properties = gaphor.services.properties.Properties(event_manager)
    new_properties.on_model_loaded(SessionCreated(None, None, "test_load_properties"))

    assert new_properties.get("test", "not set") == "not set"
    assert "Invalid syntax in property file" in caplog.text



from gaphas.segment import Segment

from gaphor.diagram.general.simpleitem import Line
from gaphor.diagram.tests.fixtures import connect
from gaphor.UML.general import CommentLineItem


def test_expose_merge_issue(diagram):
    line = diagram.create(Line)
    segment = Segment(line, diagram)
    segment.split_segment(0)

    comment_line = diagram.create(CommentLineItem)
    connect(comment_line, comment_line.head, line, line.ports()[1])

    segment.merge_segment(0)

    diagram.unlink()



from gaphas.tool.rubberband import RubberbandState

from gaphor.diagram.diagramtoolbox import new_item_factory
from gaphor.diagram.general import Box
from gaphor.diagram.tools import (
    apply_default_tool_set,
    apply_magnet_tool_set,
    apply_placement_tool_set,
)


def test_default_tool_set(view, modeling_language, event_manager):
    apply_default_tool_set(view, modeling_language, event_manager, RubberbandState())

    assert view.observe_controllers()


def test_magnet_tool_set(view, modeling_language, event_manager):
    apply_magnet_tool_set(view, modeling_language, event_manager)

    assert view.observe_controllers()


def test_placement_tool_set(view, modeling_language, event_manager):
    apply_placement_tool_set(
        view, new_item_factory(Box), modeling_language, event_manager, 0
    )

    assert view.observe_controllers()



import sys

import pytest
from gi.repository import GLib, Gtk

import gaphor.main

APP_NAME = sys.argv[0]


@pytest.fixture(autouse=True)
def custom_application_id():
    orig_app_id = gaphor.ui.APPLICATION_ID
    gaphor.ui.APPLICATION_ID = "org.gaphor.TestMain"
    yield
    gaphor.ui.APPLICATION_ID = orig_app_id


def fake_run(monkeypatch):
    run = []

    def fake_run(self, args):
        run.extend(args)
        # NB. Priority used to be GLib.PRIORITY_LOW + GLib.PRIORITY_LOW,
        # but that makes the tests hang. Is an idle loop blocking us?
        GLib.idle_add(self.quit, priority=GLib.PRIORITY_LOW)
        app_run(self, args)

    app_run = Gtk.Application.run
    monkeypatch.setattr(Gtk.Application, "run", fake_run)

    return run


def test_application_startup(monkeypatch):
    run = fake_run(monkeypatch)

    gaphor.main.main([APP_NAME])

    assert run == [APP_NAME]


def test_application_startup_with_model(monkeypatch):
    run = fake_run(monkeypatch)

    gaphor.main.main([APP_NAME, "test-models/all-elements.gaphor"])

    assert run == [APP_NAME, "test-models/all-elements.gaphor"]



from gaphor.services.componentregistry import ComponentRegistry


class Dependency:
    pass


def test_no_services_for_partial_service_application():
    component_registry = ComponentRegistry()
    dependency = Dependency()
    component_registry.register("dependency", dependency)

    class TestSubject:
        def __init__(self, myarg):
            self.myarg = myarg

    subject = component_registry.partial(TestSubject)(1)

    assert subject.myarg == 1


def test_partial_service_application():
    component_registry = ComponentRegistry()
    dependency = Dependency()
    component_registry.register("dependency", dependency)

    class TestSubject:
        def __init__(self, dependency):
            self.dependency = dependency

    subject = component_registry.partial(TestSubject)()

    assert subject.dependency is dependency



import pytest

from gaphor.services.undomanager import UndoManager


@pytest.fixture
def undo_manager(event_manager, element_factory):
    undo_manager = UndoManager(event_manager, element_factory)
    yield undo_manager
    undo_manager.shutdown()



import pytest

from gaphor.SysML.painter import SysMLDiagramTypePainter
from gaphor.SysML.sysml import Block, ConstraintBlock, Requirement, SysMLDiagram
from gaphor.UML.uml import Activity, Interaction, Package, Profile, StateMachine


@pytest.mark.parametrize(
    "type,label",
    [
        [Activity, "[activity] ElemName [DiagName]"],
        [Block, "[block] ElemName [DiagName]"],
        [ConstraintBlock, "[constraint block] ElemName [DiagName]"],
        [Interaction, "[interaction] ElemName [DiagName]"],
        [Package, "[package] ElemName [DiagName]"],
        [Profile, "[profile] ElemName [DiagName]"],
        [Requirement, "[requirement] ElemName [DiagName]"],
        [StateMachine, "[state machine] ElemName [DiagName]"],
    ],
)
def test_diagram_label(type, label):
    def _diagram(elementType):
        element = elementType()
        element.name = "ElemName"

        diagram = SysMLDiagram()
        diagram.name = "DiagName"
        diagram.element = element

        return diagram

    assert SysMLDiagramTypePainter(_diagram(type)).label() == label


def test_uncontained_diagram_label():
    diagram = SysMLDiagram()
    diagram.name = "DiagName"

    assert SysMLDiagramTypePainter(diagram).label() == "DiagName"



import pytest

from gaphor import UML
from gaphor.core.modeling import Diagram
from gaphor.diagram.tests.fixtures import allow, connect, disconnect, get_connected
from gaphor.UML.classes.association import AssociationItem
from gaphor.UML.classes.klass import ClassItem


@pytest.fixture
def connected_association(create):
    asc = create(AssociationItem)
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)

    connect(asc, asc.head, c1)
    assert asc.subject is None  # no UML metaclass yet

    connect(asc, asc.tail, c2)
    assert asc.subject is not None

    return asc, c1, c2


@pytest.fixture
def clone(create):
    def _clone(item):
        new = create(type(item))
        new.subject = item.subject
        new.head_subject = item.head_subject
        new.tail_subject = item.tail_subject
        return new

    return _clone


def test_glue_to_class(connected_association):
    asc, c1, c2 = connected_association

    glued = allow(asc, asc.head, c1)
    assert glued

    connect(asc, asc.head, c1)

    glued = allow(asc, asc.tail, c2)
    assert glued


def test_association_item_connect(connected_association, element_factory):
    asc, _c1, _c2 = connected_association

    # Diagram, Class *2, Property *2, Association
    assert len(element_factory.lselect()) == 9
    assert asc.head_subject is not None
    assert asc.tail_subject is not None


def test_association_item_reconnect_copies_properties(connected_association, create):
    asc, c1, c2 = connected_association
    c3 = create(ClassItem, UML.Class)

    asc.subject.name = "Name"

    a = asc.subject

    connect(asc, asc.tail, c3)

    assert a is not asc.subject
    ends = [p.type for p in asc.subject.memberEnd]
    assert c1.subject in ends
    assert c3.subject in ends
    assert c2.subject not in ends
    assert asc.subject.name == "Name"


def test_association_item_reconnect_with_navigability(connected_association, create):
    asc, _c1, _c2 = connected_association
    c3 = create(ClassItem, UML.Class)

    UML.recipes.set_navigability(asc.subject, asc.tail_subject, True)
    connect(asc, asc.tail, c3)

    assert asc.tail_subject.navigability is True


def test_association_item_reconnect_with_aggregation(connected_association, create):
    asc, _c1, _c2 = connected_association
    c3 = create(ClassItem, UML.Class)

    asc.tail_subject.aggregation = "composite"
    connect(asc, asc.tail, c3)

    assert asc.tail_subject.aggregation == "composite"


def test_disconnect_should_disconnect_model(
    connected_association, element_factory, sanitizer_service
):
    asc, c1, c2 = connected_association

    disconnect(asc, asc.head)
    disconnect(asc, asc.tail)
    assert c1 is not get_connected(asc, asc.head)
    assert c2 is not get_connected(asc, asc.tail)

    assert not asc.subject
    assert not asc.head_subject
    assert not asc.tail_subject
    assert not element_factory.lselect(UML.Property)


def test_disconnect_of_second_association_should_leave_model_in_tact(
    connected_association, clone, sanitizer_service
):
    asc, c1, c2 = connected_association
    new = clone(asc)

    disconnect(new, new.head)
    assert asc.subject.memberEnd[0].type is c1.subject
    assert asc.subject.memberEnd[1].type is c2.subject
    assert new.subject is asc.subject


def test_disconnect_of_navigable_end_should_remove_owner_relationship(
    connected_association, element_factory, sanitizer_service
):
    asc, _c1, c2 = connected_association

    UML.recipes.set_navigability(asc.subject, asc.head_subject, True)

    assert asc.head_subject in c2.subject.ownedAttribute

    disconnect(asc, asc.head)

    assert not asc.subject
    assert not asc.head_subject
    assert not asc.tail_subject
    assert not element_factory.lselect(UML.Property)


def test_allow_reconnect_for_single_presentation(connected_association, create):
    asc, _c1, _c2 = connected_association
    c3 = create(ClassItem, UML.Class)

    assert allow(asc, asc.head, c3)


def test_allow_reconnect_on_same_class_for_multiple_presentations(
    connected_association, clone, create
):
    asc, c1, c2 = connected_association
    new = clone(asc)

    assert allow(new, new.head, c1)
    assert allow(new, new.tail, c2)


def test_allow_reconnect_if_only_one_connected_presentations(
    connected_association, clone, create
):
    asc, _c1, _c2 = connected_association
    clone(asc)

    c3 = create(ClassItem, UML.Class)

    assert allow(asc, asc.head, c3)


def test_create_association_in_new_diagram_should_reuse_existing(
    connected_association, element_factory
):
    asc, c1, c2 = connected_association

    diagram2 = element_factory.create(Diagram)
    c3 = diagram2.create(ClassItem, subject=c1.subject)
    c4 = diagram2.create(ClassItem, subject=c2.subject)
    asc2 = diagram2.create(AssociationItem)

    connect(asc2, asc2.head, c3)
    connect(asc2, asc2.tail, c4)

    assert asc.subject is asc2.subject
    assert asc.head_subject is asc2.head_subject
    assert asc.tail_subject is asc2.tail_subject


def test_create_association_in_new_diagram_reversed_should_reuse_existing(
    connected_association, element_factory
):
    asc, c1, c2 = connected_association

    diagram2 = element_factory.create(Diagram)
    c3 = diagram2.create(ClassItem, subject=c1.subject)
    c4 = diagram2.create(ClassItem, subject=c2.subject)
    asc2 = diagram2.create(AssociationItem)

    connect(asc2, asc2.tail, c3)
    connect(asc2, asc2.head, c4)

    assert asc.subject is asc2.subject
    assert asc.head_subject is asc2.tail_subject
    assert asc.tail_subject is asc2.head_subject


def test_disconnect_association_in_new_diagram_should_clear_ends(
    connected_association, element_factory
):
    asc, c1, c2 = connected_association

    diagram2 = element_factory.create(Diagram)
    c3 = diagram2.create(ClassItem, subject=c1.subject)
    c4 = diagram2.create(ClassItem, subject=c2.subject)
    asc2 = diagram2.create(AssociationItem)

    connect(asc2, asc2.tail, c3)
    connect(asc2, asc2.head, c4)
    disconnect(asc, asc.head)

    assert not asc.subject
    assert not asc.head_subject
    assert not asc.tail_subject


@pytest.mark.xfail(
    reason="We cannot determine if the head of tail end of the association is missing"
)
# See https://github.com/gaphor/gaphor/pull/2741
def test_connect_association_missing_head_subject(create, element_factory):
    asc_item = create(AssociationItem, UML.Association)
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)

    asc = asc_item.subject = UML.recipes.create_association(c1.subject, c2.subject)

    # Now break the association by removing one end:
    asc.memberEnd[0].unlink()

    connect(asc_item, asc_item.tail, c1)
    connect(asc_item, asc_item.head, c2)

    assert len(asc_item.subject.memberEnd) == 2
    assert asc_item.head_subject
    assert asc_item.tail_subject



import pytest

from gaphor import UML
from gaphor.diagram.copypaste import copy_full, paste_link
from gaphor.diagram.group import group
from gaphor.diagram.tests.fixtures import copy_clear_and_paste_link
from gaphor.UML.deployments import ArtifactItem, NodeItem


@pytest.fixture
def node_with_artifact(diagram, element_factory):
    node = element_factory.create(UML.Node)
    artifact = element_factory.create(UML.Artifact)
    node_item = diagram.create(NodeItem, subject=node)
    artifact_item = diagram.create(ArtifactItem, subject=artifact)

    group(node_item, artifact_item)
    artifact_item.change_parent(node_item)

    assert artifact_item.parent is node_item

    return node_item, artifact_item


def test_copy_paste_of_nested_item(diagram, node_with_artifact):
    node_item, artifact_item = node_with_artifact

    buffer = copy_full({artifact_item})

    (new_comp_item,) = paste_link(buffer, diagram)

    assert new_comp_item.parent is node_item


def test_copy_paste_of_item_with_nested_item(diagram, node_with_artifact):
    buffer = copy_full(set(node_with_artifact))

    new_items = paste_link(buffer, diagram)

    new_node_item = next(i for i in new_items if isinstance(i, NodeItem))
    new_comp_item = next(i for i in new_items if isinstance(i, ArtifactItem))

    assert new_comp_item.parent is new_node_item


def test_copy_remove_paste_of_item_with_nested_item(
    diagram, element_factory, node_with_artifact
):
    new_items = copy_clear_and_paste_link(
        set(node_with_artifact), diagram, element_factory
    )

    new_node_item = next(i for i in new_items if isinstance(i, NodeItem))
    new_comp_item = next(i for i in new_items if isinstance(i, ArtifactItem))

    assert new_comp_item.parent is new_node_item



import pytest
from gaphas import Item

from gaphor.core.modeling import Diagram
from gaphor.diagram.tests.fixtures import connect
from gaphor.SysML.blocks.block import BlockItem
from gaphor.SysML.blocks.proxyport import ProxyPortItem
from gaphor.SysML.sysml import Block


def test_proxy_port_item_conforms_to_item_protocol(diagram):
    item = ProxyPortItem(diagram)

    assert isinstance(item, Item)


def test_proxy_port_item_point(diagram):
    item = ProxyPortItem(diagram)

    assert item.point(0, 0) == pytest.approx(0.0)


def test_ports(diagram):
    proxy_port = diagram.create(ProxyPortItem)

    diagram.update({proxy_port})

    top, right, bottom, left = proxy_port.ports()

    def pos(p):
        return p.x, p.y

    assert pos(proxy_port.handles()[0].pos) == (0, 0)
    assert pos(left.end) == pos(top.start) == (-8, -8)
    assert pos(top.end) == pos(right.start) == (8, -8)
    assert pos(right.end) == pos(bottom.start) == (8, 8)
    assert pos(bottom.end) == pos(left.start) == (-8, 8)


def test_save_load(create, saver, loader, element_factory):
    proxy_port = create(ProxyPortItem)
    block = create(BlockItem, Block)

    connect(proxy_port, proxy_port.handles()[0], block)

    data = saver()
    loader(data)

    new_diagram = next(element_factory.select(Diagram))
    new_proxy_port = next(element_factory.select(ProxyPortItem))
    new_block = next(element_factory.select(BlockItem))
    connected = new_diagram.connections.get_connection(
        new_proxy_port.handles()[0]
    ).connected

    assert connected is new_block



import pytest

from gaphor.core.modeling.presentation import literal_eval


@pytest.mark.parametrize(
    "text,expected",
    [
        ("12", 12),
        ("  12", 12),
        ("12  ", 12),
        ("12\n  ", 12),
        (
            """[
        1, 2,
        3
        ]
        """,
            [1, 2, 3],
        ),
        (
            """
        [
        1, 2,
        3
        ]
        """,
            [1, 2, 3],
        ),
    ],
)
def test_literal_eval(text, expected):
    assert literal_eval(text) == expected



import itertools

import pytest

from gaphor.services.modelinglanguage import ModelingLanguageService
from gaphor.storage.load import (
    load_elements,
    sysml_diagram_type_to_class,
    uml_diagram_type_to_class,
)
from gaphor.storage.parser import element
from gaphor.storage.upgrade_canvasitem import upgrade_canvasitem
from gaphor.UML import Dependency, Diagram, Image, Package, diagramitems


@pytest.fixture
def loader(element_factory, modeling_language):
    def _loader(*parsed_items):
        for item in parsed_items:
            if item.type.endswith("Item"):
                item.references["diagram"] = "1"
                upgrade_canvasitem(item, "1.0.0")
        parsed_data = {
            "1": element(id="1", type="Diagram"),
            **{p.id: p for p in parsed_items},
        }
        load_elements(parsed_data, element_factory, modeling_language)
        return list(element_factory.lselect()[0].get_all_items())

    return _loader


def test_upgrade_metaclass_item_to_class_item(loader):
    item = loader(element(id="2", type="MetaclassItem"))[0]

    assert type(item) is diagramitems.ClassItem


def test_upgrade_subsystem_item_to_class_item(loader):
    item = loader(element(id="2", type="SubsystemItem"))[0]

    assert type(item) is diagramitems.ComponentItem


def test_rename_stereotype_attrs_field(loader):
    parsed_item = element(id="2", type="ClassItem")
    parsed_item.values["show_stereotypes_attrs"] = "1"
    item = loader(parsed_item)[0]

    assert not hasattr(item, "show_stereotypes_attrs")
    assert item.show_stereotypes


def test_rename_show_attributes_and_operations_field(loader):
    parsed_item = element(id="2", type="ClassItem")
    parsed_item.values["show-attributes"] = "0"
    parsed_item.values["show-operations"] = "0"
    item = loader(parsed_item)[0]

    assert not item.show_attributes
    assert not item.show_operations


def test_interface_drawing_style_normal(loader):
    parsed_item = element(id="2", type="InterfaceItem")
    parsed_item.values["drawing-style"] = "0"  # DRAW_NONE
    item = loader(parsed_item)[0]

    assert item.folded.name == "NONE"


def test_interface_drawing_style_folded(loader):
    parsed_item = element(id="2", type="InterfaceItem")
    parsed_item.values["drawing-style"] = "3"  # DRAW_ICON
    item = loader(parsed_item)[0]

    assert item.folded.name == "PROVIDED"


def test_upgrade_generalization_arrow_direction(loader):
    cls1 = element(id="2", type="ClassItem")
    cls2 = element(id="3", type="ClassItem")
    gen_item = element(id="4", type="GeneralizationItem")
    gen_item.references["head-connection"] = ["2"]
    gen_item.references["tail-connection"] = ["3"]

    cls1, cls2, gen_item = loader(cls1, cls2, gen_item)

    assert gen_item.diagram.connections.get_connection(gen_item.head).connected is cls2
    assert gen_item.diagram.connections.get_connection(gen_item.tail).connected is cls1


def test_upgrade_flow_item_to_control_flow_item(element_factory, modeling_language):
    diagram = element(id="1", type="Diagram")
    objnode = element(id="2", type="ControlFlow")
    item = element(id="3", type="FlowItem")
    item.references["diagram"] = diagram.id
    item.references["subject"] = objnode.id

    load_elements(
        {p.id: p for p in (diagram, objnode, item)}, element_factory, modeling_language
    )

    assert element_factory.lselect(diagramitems.ControlFlowItem)


def test_upgrade_flow_item_to_object_flow_item(element_factory, modeling_language):
    diagram = element(id="1", type="Diagram")
    objnode = element(id="2", type="ObjectFlow")
    item = element(id="3", type="FlowItem")
    item.references["diagram"] = diagram.id
    item.references["subject"] = objnode.id

    load_elements(
        {p.id: p for p in (diagram, objnode, item)}, element_factory, modeling_language
    )

    assert element_factory.lselect(diagramitems.ObjectFlowItem)


def test_upgrade_decision_node_item_show_type(loader):
    parsed_item = element(id="2", type="DecisionNodeItem")
    parsed_item.values["show_type"] = "1"
    item = loader(parsed_item)[0]

    assert item.show_underlying_type == 1


@pytest.mark.parametrize("type", ["Property", "Port", "ProxyPort"])
def test_upgrade_delete_property_information_flow(
    element_factory, modeling_language, type
):
    prop = element(id="1", type=type)
    iflow = element(id="2", type="InformationFlow")
    prop.references["informationFlow"] = ["1"]

    load_elements({e.id: e for e in (prop, iflow)}, element_factory, modeling_language)

    assert element_factory.lselect(modeling_language.lookup_element(type))
    assert element_factory.lselect(modeling_language.lookup_element("InformationFlow"))


def test_upgrade_note_on_model_element(loader, element_factory):
    cls_item = element(id="2", type="ClassItem")
    cls = element(id="3", type="Class")
    cls_item.values["note"] = "my note"
    cls_item.references["subject"] = cls.id

    loader(cls_item, cls)
    _, cls_item, cls, *_ = element_factory.lselect()

    assert cls.note == "my note"


def test_upgrade_append_notes_on_model_element(loader, element_factory):
    cls_item1 = element(id="2", type="ClassItem")
    cls_item2 = element(id="3", type="ClassItem")
    cls = element(id="44", type="Class")
    cls_item1.values["note"] = "my note"
    cls_item1.references["subject"] = cls.id
    cls_item2.values["note"] = "another note"
    cls_item2.references["subject"] = cls.id

    loader(cls_item1, cls_item2, cls)
    _, cls_item1, cls_item2, cls, *_ = element_factory.lselect()

    assert cls.note == "my note\n\nanother note"


def test_upgrade_picture_to_image(loader, element_factory):
    picture = element(id="2", type="Picture")
    picture_item = element(id="3", type="PictureItem")

    loader(picture, picture_item)
    _, image, image_item, *_ = element_factory.lselect()

    assert isinstance(image, Image)
    assert isinstance(image_item, diagramitems.ImageItem)


def test_upgrade_diagram_to_uml_diagram(loader, element_factory):
    diagram = element(id="2", type="Diagram")
    core_diagram = element(id="3", type="Diagram", ns="Core")

    loader(diagram, core_diagram)
    _, new_diagram, new_core_diagram, *_ = element_factory.lselect()

    assert isinstance(new_diagram, Diagram)
    assert isinstance(new_core_diagram, Diagram)


def test_upgrade_dependency_owning_package(loader, element_factory):
    pkg = element(id="2", type="Package")
    pkg.references["ownedDiagram"] = ["1"]
    cls_item1 = element(id="3", type="ClassItem")
    cls1 = element(id="30", type="Class")
    cls_item1.references["subject"] = cls1.id
    cls_item2 = element(id="4", type="ClassItem")
    cls2 = element(id="40", type="Class")
    cls_item2.references["subject"] = cls2.id
    dep_item = element(id="5", type="DependencyItem", ns="UML")
    dep = element(id="50", type="Dependency", ns="UML")
    dep_item.references["subject"] = dep.id
    dep.references["supplier"] = [cls1.id]
    dep.references["client"] = [cls2.id]

    loader(cls_item1, cls_item2, dep_item, pkg, dep, cls1, cls2)
    new_pkg = next(element_factory.select(Package))
    new_dep = next(element_factory.select(Dependency))

    assert new_dep.owner is new_pkg


@pytest.mark.parametrize(
    "orig_type,kind,ns,new_type",
    [
        ["Diagram", "", "UML", "Diagram"],
        ["Diagram", "cls", "UML", "ClassDiagram"],
        ["Diagram", "pkg", "UML", "PackageDiagram"],
        ["Diagram", "cmp", "UML", "ComponentDiagram"],
        ["Diagram", "dep", "UML", "DeploymentDiagram"],
        ["Diagram", "prf", "UML", "ProfileDiagram"],
        ["Diagram", "act", "UML", "ActivityDiagram"],
        ["Diagram", "sd", "UML", "SequenceDiagram"],
        ["Diagram", "com", "UML", "CommunicationDiagram"],
        ["Diagram", "stm", "UML", "StateMachineDiagram"],
        ["Diagram", "uc", "UML", "UseCaseDiagram"],
        ["Diagram", "bdd", "SysML", "BlockDefinitionDiagram"],
        ["Diagram", "bdd", "SysML", "BlockDefinitionDiagram"],
        ["Diagram", "ibd", "SysML", "InternalBlockDiagram"],
        ["Diagram", "req", "SysML", "RequirementDiagram"],
        ["SysMLDiagram", "bdd", "SysML", "BlockDefinitionDiagram"],
        ["SysMLDiagram", "ibd", "SysML", "InternalBlockDiagram"],
        ["SysMLDiagram", "par", "SysML", "ParametricDiagram"],
        ["SysMLDiagram", "pkg", "SysML", "PackageDiagram"],
        ["SysMLDiagram", "req", "SysML", "RequirementDiagram"],
        ["SysMLDiagram", "act", "SysML", "ActivityDiagram"],
        ["SysMLDiagram", "sd", "SysML", "SequenceDiagram"],
        ["SysMLDiagram", "stm", "SysML", "StateMachineDiagram"],
        ["SysMLDiagram", "uc", "SysML", "UseCaseDiagram"],
    ],
)
def test_diagram_type_to_class(
    orig_type, kind, ns, new_type, element_factory, modeling_language
):
    e = element(id="1", type=orig_type)
    e.values["diagramType"] = kind

    load_elements({"1": e}, element_factory, modeling_language)
    diagram = next(element_factory.select())
    expected_type = modeling_language.lookup_element(new_type, ns)

    assert type(diagram) is expected_type


@pytest.mark.parametrize(
    "kind,ns,name",
    (
        [(k, n, m) for k, (n, m) in uml_diagram_type_to_class.items()]
        + [(k, n, m) for k, (n, m) in sysml_diagram_type_to_class.items()]
    ),
)
def test_diagram_type_to_class_mapping(kind, ns, name):
    modeling_language = ModelingLanguageService()
    diagram_class = modeling_language.lookup_element(name, ns)

    assert diagram_class.diagramType.default == kind


@pytest.mark.parametrize(
    "element_type,property_name,value,literal_type,expected_value",
    [
        ["MultiplicityElement", "lowerValue", "1", "LiteralInteger", 1],
        ["MultiplicityElement", "lowerValue", "", "NoneType", None],
        ["MultiplicityElement", "lowerValue", None, "NoneType", None],
        ["MultiplicityElement", "upperValue", "1", "LiteralUnlimitedNatural", 1],
        ["MultiplicityElement", "upperValue", "*", "LiteralUnlimitedNatural", "*"],
        ["MultiplicityElement", "upperValue", "", "LiteralUnlimitedNatural", "*"],
        ["MultiplicityElement", "upperValue", None, "NoneType", None],
        *(
            a[0] + a[1]
            for a in itertools.product(
                [
                    ["Parameter", "defaultValue"],
                    ["Slot", "value"],
                    ["ValueSpecificationAction", "value"],
                    ["InstanceSpecification", "specification"],
                    ["Constraint", "specification"],
                    ["ActivityEdge", "guard"],
                    ["ControlFlow", "guard"],
                    ["ObjectFlow", "guard"],
                    ["JoinNode", "joinSpec"],
                ],
                [
                    ["value", "LiteralString", "value"],
                    ['"value"', "LiteralString", "value"],
                    ["1", "LiteralInteger", 1],
                    ["*", "LiteralUnlimitedNatural", "*"],
                    ["true", "LiteralBoolean", True],
                    ["True", "LiteralBoolean", True],
                    ["false", "LiteralBoolean", False],
                    ["False", "LiteralBoolean", False],
                    ["", "LiteralString", ""],
                    [None, "LiteralString", None],
                    ['"', "LiteralString", ""],
                    ["-1", "LiteralString", "-1"],
                    ["1a", "LiteralString", "1a"],
                ],
            )
        ),
    ],
)
def test_update_value_to_value_specification(
    element_type,
    property_name,
    value,
    literal_type,
    expected_value,
    element_factory,
    modeling_language,
):
    e = element(id="1", type=element_type)
    e.values[property_name] = value

    load_elements({"1": e}, element_factory, modeling_language)

    p = element_factory.lookup("1")
    value_spec = getattr(p, property_name)

    assert type(p).__name__ == element_type
    assert type(value_spec).__name__ == literal_type

    if value_spec is not None:
        assert value_spec.value == expected_value
        assert value_spec in element_factory


@pytest.mark.parametrize(
    "element_type",
    [
        "LiteralString",
        "LiteralInteger",
        "LiteralBoolean",
        "LiteralUnlimitedNatural",
        "LiteralSpecification",
    ],
)
def test_do_not_update_value_for_literal_specification(
    element_type, element_factory, modeling_language
):
    e = element(id="1", type=element_type)
    e.values["value"] = "1"

    load_elements({"1": e}, element_factory, modeling_language)

    p = element_factory.lookup("1")

    assert type(p).__name__ == element_type
    assert str(p.value) == "1"


def test_update_parameter_default_direction(element_factory, modeling_language):
    e = element(id="2", type="Parameter")
    e.values["direction"] = "out"

    load_elements(
        {"1": element(id="1", type="Parameter"), "2": e},
        element_factory,
        modeling_language,
    )

    p1 = element_factory.lookup("1")
    p2 = element_factory.lookup("2")

    assert p1.direction == "inout"
    assert p2.direction == "out"


@pytest.mark.parametrize(
    "element_type",
    [
        "ObjectNode",
        "ActivityParameterNode",
        "Pin",
        "InputPin",
        "OutputPin",
    ],
)
def test_update_object_node_default_ordering(
    element_type, element_factory, modeling_language
):
    load_elements(
        {"1": element(id="1", type=element_type)}, element_factory, modeling_language
    )

    p = element_factory.lookup("1")

    assert p.ordering == "unordered"



"""Tests for the HTML report generator."""

import pytest

from gaphor import UML
from gaphor.plugins.htmlreport.generator import (
    _ICON_FA_CLASSES,
    _icon_fa_class,
    build_model_data,
    build_tree,
    extract_element_info,
    generate_report,
)
from gaphor.UML.classes import ClassItem
from gaphor.UML.recipes import apply_stereotype


@pytest.fixture
def model_with_diagram(diagram, element_factory, create):
    c1 = create(ClassItem, UML.Class)
    c1.subject.name = "MyClass"
    c1.matrix.translate(50, 50)

    diagram.name = "Test Diagram"
    diagram.update(diagram.ownedPresentation)

    return diagram, element_factory


def test_index_html_contains_model_data(model_with_diagram, tmp_path):
    diagram, factory = model_with_diagram
    output_dir = tmp_path / "report"
    generate_report(factory, output_dir)

    html = (output_dir / "index.html").read_text(encoding="utf-8")
    assert "MODEL_DATA" in html
    assert "MyClass" in html
    assert "Test Diagram" in html


def test_build_model_data_structure(model_with_diagram):
    diagram, factory = model_with_diagram
    data = build_model_data(factory)

    assert "elements" in data
    assert "diagrams" in data
    assert "tree" in data

    # Check diagram is present
    assert len(data["diagrams"]) >= 1
    diagram_entry = next(iter(data["diagrams"].values()))
    assert diagram_entry["name"] == "Test Diagram"
    assert "svg_content" in diagram_entry

    # Check element is present
    found_class = False
    for el in data["elements"].values():
        if el["name"] == "MyClass":
            found_class = True
            assert el["type"] == "Class"
            assert len(el["diagrams"]) >= 1
    assert found_class


def test_extract_element_info_with_abstract(element_factory):
    cls = element_factory.create(UML.Class)
    cls.name = "AbstractClass"
    cls.isAbstract = True

    info = extract_element_info(cls)
    assert info is not None
    props = {p["name"]: p["value"] for p in info["properties"]}
    assert "isAbstract" in props


def test_extract_element_info_association(diagram, element_factory, create):
    from gaphor.diagram.tests.fixtures import connect
    from gaphor.UML.classes.association import AssociationItem

    c1 = create(ClassItem, UML.Class)
    c1.subject.name = "Src"
    c2 = create(ClassItem, UML.Class)
    c2.subject.name = "Dst"

    assoc = create(AssociationItem)
    connect(assoc, assoc.head, c1)
    connect(assoc, assoc.tail, c2)

    diagram.update(diagram.ownedPresentation)

    info = extract_element_info(assoc.subject)
    assert info is not None
    assert info["type"] == "Association"
    assert len(info["associations"]) == 2
    types = {a["type"] for a in info["associations"]}
    assert "Src" in types
    assert "Dst" in types


def test_extract_element_info_generalization(element_factory):
    parent = element_factory.create(UML.Class)
    parent.name = "Base"
    child = element_factory.create(UML.Class)
    child.name = "Derived"
    gen = element_factory.create(UML.Generalization)
    gen.general = parent
    gen.specific = child

    info = extract_element_info(child)
    assert len(info["generalizations"]) == 1
    assert info["generalizations"][0]["name"] == "Base"
    assert info["generalizations"][0]["type"] == "Class"


def test_extract_element_info_stereotype(element_factory):
    cls = element_factory.create(UML.Class)
    cls.name = "Controller"
    stereotype = element_factory.create(UML.Stereotype)
    stereotype.name = "service"
    apply_stereotype(cls, stereotype)

    info = extract_element_info(cls)
    assert "service" in info["stereotypes"]


def test_extract_element_info_enumeration(element_factory):
    enum = element_factory.create(UML.Enumeration)
    enum.name = "Color"
    for name in ("Red", "Green", "Blue"):
        lit = element_factory.create(UML.EnumerationLiteral)
        lit.name = name
        enum.ownedLiteral = lit

    info = extract_element_info(enum)
    assert info["type"] == "Enumeration"
    assert len(info["literals"]) == 3
    assert "Red" in info["literals"]
    assert "Green" in info["literals"]
    assert "Blue" in info["literals"]


def test_build_tree_nested_packages(element_factory):
    pkg = element_factory.create(UML.Package)
    pkg.name = "MyPackage"
    cls = element_factory.create(UML.Class)
    cls.name = "Inner"
    cls.package = pkg

    tree = build_tree(element_factory)
    pkg_node = next(n for n in tree if n["name"] == "MyPackage")
    assert any(c["name"] == "Inner" for c in pkg_node["children"])


def test_build_tree_node_has_icon(element_factory):
    cls = element_factory.create(UML.Class)
    cls.name = "Foo"
    tree = build_tree(element_factory)
    node = next(n for n in tree if n["name"] == "Foo")
    assert node["icon"] == "fa-solid fa-c"


def test_icon_mapping_covers_common_types(element_factory):
    for element_type in (
        UML.Class,
        UML.Package,
        UML.Interface,
        UML.Component,
        UML.Enumeration,
        UML.Actor,
        UML.UseCase,
        UML.Activity,
        UML.StateMachine,
        UML.Artifact,
    ):
        el = element_factory.create(element_type)
        assert _icon_fa_class(el) != "fa-solid fa-circle", (
            f"{element_type.__name__} should have a specific icon mapping"
        )


def test_icon_fa_classes_values_are_valid():
    for icon_name, fa_class in _ICON_FA_CLASSES.items():
        parts = fa_class.split()
        assert len(parts) == 2, f"{icon_name}: expected 2 parts, got {parts}"
        assert parts[0] in ("fa-solid", "fa-regular", "fa-brands"), (
            f"{icon_name}: unexpected prefix {parts[0]}"
        )
        assert parts[1].startswith("fa-"), (
            f"{icon_name}: icon name should start with fa-"
        )



from pathlib import Path

import pytest

from gaphor.application import Application
from gaphor.event import ModelSaved, SessionCreated


@pytest.fixture
def application():
    application = Application()
    yield application
    application.shutdown()


@pytest.mark.asyncio
async def test_service_load(application):
    """Test loading services and querying utilities."""

    session = application.new_session()

    assert session.get_service("undo_manager") is not None, (
        "Failed to load the undo manager service"
    )

    assert session.get_service("file_manager") is not None, (
        "Failed to load the file manager service"
    )


@pytest.mark.asyncio
async def test_model_loaded(application, test_models):
    session = application.new_session()
    session.event_manager.handle(
        SessionCreated(None, session, filename=test_models / "all-elements.gaphor")
    )

    assert session.filename == test_models / "all-elements.gaphor"


@pytest.mark.asyncio
async def test_model_saved(application):
    session = application.new_session()
    session.event_manager.handle(ModelSaved(Path("some_file_name")))

    assert session.filename == Path("some_file_name")


@pytest.mark.asyncio
async def test_new_session_from_template(application, test_models, event_manager):
    with (test_models / "test-model.gaphor").open(encoding="utf-8") as model:
        session = application.new_session(template=model)
        await session.get_service("event_manager").gather_tasks()

    assert any(session.get_service("element_factory").select())



import pytest

from gaphor.C4Model.modelinglanguage import C4ModelLanguage


def test_modeling_language_name():
    ml = C4ModelLanguage()
    assert ml.name


@pytest.mark.parametrize("name", ["Person", "Container"])
def test_elements(name):
    ml = C4ModelLanguage()

    assert ml.lookup_element(name)


@pytest.mark.parametrize(
    "name",
    [
        "C4PersonItem",
        "C4ContainerItem",
        "C4DatabaseItem",
    ],
)
def test_diagram_items(name):
    ml = C4ModelLanguage()

    assert ml.lookup_element(name)



from gaphas.item import Item

from gaphor.core.eventmanager import event_handler
from gaphor.core.modeling.diagram import Diagram
from gaphor.core.modeling.event import ElementDeleted
from gaphor.core.modeling.presentation import Presentation
from gaphor.tests.raises import raises_exception_group


class Example(Presentation, Item):
    pass


def test_presentation_implements_item_protocol(diagram):
    presentation = diagram.create(Example)

    assert isinstance(presentation, Item)


def test_presentation_should_have_a_diagram(diagram):
    presentation = diagram.create(Example)

    assert presentation.diagram is diagram


def test_should_emit_event_when_unlinked(diagram, event_manager):
    presentation = diagram.create(Example)
    events = []

    @event_handler(ElementDeleted)
    def handler(event):
        events.append(event)

    event_manager.subscribe(handler)

    presentation.unlink()

    assert events
    assert events[0].diagram is diagram
    assert events[0].element is presentation


def test_should_unlink_more_than_once(diagram, event_manager):
    presentation = diagram.create(Example)
    events = []

    @event_handler(ElementDeleted)
    def handler(event):
        events.append(event)

    event_manager.subscribe(handler)

    presentation.unlink()
    presentation.unlink()
    presentation.unlink()

    assert len(events) == 1


def test_presentation_can_not_set_new_diagram(diagram, element_factory):
    presentation = diagram.create(Example)
    new_diagram = element_factory.create(Diagram)

    with raises_exception_group(ValueError):
        presentation.diagram = new_diagram

    assert presentation.diagram is diagram


def test_matrix_i2c_updates_when_matrix_changes(diagram):
    presentation = diagram.create(Example)

    presentation.matrix.translate(1, 1)

    assert tuple(presentation.matrix_i2c) == (1, 0, 0, 1, 1, 1)


def test_parent_matrix_updates(diagram):
    parent = diagram.create(Example)
    presentation = diagram.create(Example)

    presentation.change_parent(parent)
    parent.matrix.scale(2, 2)

    assert tuple(presentation.matrix_i2c) == (2, 0, 0, 2, 0, 0)


def test_set_parent_updates_matrix_i2c(diagram):
    parent = diagram.create(Example)
    presentation = diagram.create(Example)

    parent.matrix.scale(2, 2)
    presentation.change_parent(parent)

    assert tuple(presentation.matrix_i2c) == (1, 0, 0, 1, 0, 0)


def test_unset_parent_updates_matrix_i2c(diagram):
    parent = diagram.create(Example)
    presentation = diagram.create(Example)

    parent.matrix.scale(2, 2)
    presentation.change_parent(parent)
    presentation.change_parent(None)

    assert tuple(presentation.matrix_i2c) == (1, 0, 0, 1, 0, 0)


def test_change_parent_updates_matrix_i2c_and_keeps_coordinates(diagram):
    parent = diagram.create(Example)
    new_parent = diagram.create(Example)
    presentation = diagram.create(Example)

    parent.matrix.scale(2, 2)
    new_parent.matrix.translate(2, 2)
    presentation.change_parent(parent)
    presentation.change_parent(new_parent)

    assert tuple(presentation.matrix_i2c) == (1, 0, 0, 1, 0, 0)



import pytest
from gi.repository import GLib, Gtk

import gaphor.services.componentregistry
import gaphor.ui.menufragment
from gaphor.core.modeling import ElementFactory
from gaphor.plugins.console.consolewindow import ConsoleWindow


class MainWindowStub:
    def __init__(self):
        self.window = None


@pytest.fixture
def component_registry():
    component_registry = gaphor.services.componentregistry.ComponentRegistry()
    component_registry.register("element_factory", ElementFactory())
    return component_registry


@pytest.fixture
def main_window():
    return MainWindowStub()


@pytest.fixture
def tools_menu():
    return gaphor.ui.menufragment.MenuFragment()


def test_open_close(component_registry, main_window, tools_menu):
    def on_activate(app):
        def auto_close():
            window.close()
            app.quit()

        window = ConsoleWindow(component_registry, main_window, tools_menu)
        window.open()
        app.add_window(window)

        idle = GLib.Idle(GLib.PRIORITY_LOW)
        idle.set_callback(auto_close)
        idle.attach()

    app = Gtk.Application(application_id="org.gaphor.tests.Console")
    app.connect("activate", on_activate)
    app.run()

    assert tools_menu.menu.get_n_items() == 1



"""Unittest the storage and parser modules."""

import re
from io import StringIO

import pytest

import gaphor.storage as storage
from gaphor import UML
from gaphor.core.modeling import Diagram, StyleSheet
from gaphor.diagram.tests.fixtures import connect
from gaphor.storage.load import version_lower_than
from gaphor.UML.classes import AssociationItem, ClassItem, InterfaceItem
from gaphor.UML.general import CommentItem


class PseudoFile:
    def __init__(self):
        self.data = ""

    def write(self, data):
        self.data += data

    def close(self):
        pass


def test_version_check():
    assert version_lower_than("0.3.0", (0, 15, 0))
    assert version_lower_than("0", (0, 15, 0))
    assert version_lower_than("0.14", (0, 15, 0))
    assert version_lower_than("0.14.1111", (0, 15, 0))
    assert not version_lower_than("0.15.0", (0, 15, 0))
    assert not version_lower_than("1.33.0", (0, 15, 0))
    assert not version_lower_than("0.15.0b123", (0, 15, 0))
    assert version_lower_than("0.14.0.b1", (0, 15, 0))
    assert not version_lower_than("0.15.b1", (0, 15, 0))
    assert not version_lower_than("0.16.b1", (0, 15, 0))
    assert not version_lower_than("0.15.0.b2", (0, 14, 99))
    assert not version_lower_than("1.2.0rc2-dev0+7fad31a0", (0, 17, 0))


def test_save_uml(element_factory):
    """Saving gaphor.UML model elements."""
    element_factory.create(UML.Package)
    element_factory.create(Diagram)
    element_factory.create(UML.Comment)
    element_factory.create(UML.Class)

    out = PseudoFile()
    storage.save(out, element_factory=element_factory)
    out.close()

    assert "<UML:Package " in out.data
    assert "<Core:Diagram " in out.data
    assert "<UML:Comment " in out.data
    assert "<UML:Class " in out.data


def test_save_item(diagram, element_factory):
    """Save a diagram item too."""
    diagram = element_factory.create(Diagram)
    diagram.create(CommentItem, subject=element_factory.create(UML.Comment))

    out = PseudoFile()
    storage.save(out, element_factory=element_factory)
    out.close()

    assert "<Core:Diagram " in out.data
    assert "<UML:Comment " in out.data
    assert "<canvas>" not in out.data
    assert "<UML:CommentItem " in out.data, out.data


def test_load_uml(element_factory, saver, loader):
    """Test loading of a freshly saved model."""
    element_factory.create(UML.Package)
    # diagram is created in case's init
    element_factory.create(UML.Comment)
    element_factory.create(UML.Class)

    data = saver()
    loader(data)

    assert len(element_factory.lselect()) == 4
    assert len(element_factory.lselect(UML.Package)) == 1
    assert len(element_factory.lselect(UML.Comment)) == 1
    assert len(element_factory.lselect(UML.Class)) == 1
    assert len(element_factory.lselect(StyleSheet)) == 1


def test_load_uml_2(create, element_factory, saver, loader):
    """Test loading of a freshly saved model."""
    element_factory.create(UML.Package)
    create(CommentItem, UML.Comment)
    create(ClassItem, UML.Class)
    iface = create(InterfaceItem, UML.Interface)
    iface.subject.name = "Circus"
    iface.matrix.translate(10, 10)

    data = saver()
    loader(data)

    assert len(element_factory.lselect()) == 9
    assert len(element_factory.lselect(UML.Package)) == 1
    assert len(element_factory.lselect(Diagram)) == 1
    d = element_factory.lselect(Diagram)[0]
    assert len(element_factory.lselect(UML.Comment)) == 1
    assert len(element_factory.lselect(UML.Class)) == 1
    assert len(element_factory.lselect(UML.Interface)) == 1

    c = element_factory.lselect(UML.Class)[0]
    assert c.presentation
    assert c.presentation[0].subject is c

    iface = element_factory.lselect(UML.Interface)[0]
    assert iface.name == "Circus"
    assert len(iface.presentation) == 1
    assert tuple(iface.presentation[0].matrix) == (1, 0, 0, 1, 10, 10), tuple(
        iface.presentation[0].matrix
    )

    # Check load/save of other canvas items.
    assert len(list(d.get_all_items())) == 3
    for item in d.get_all_items():
        assert item.subject, f"No subject for {item}"
    d1 = next(d.select(lambda e: isinstance(e, ClassItem)))
    assert d1


def test_load_with_whitespace_name(diagram, element_factory, saver, loader):
    difficult_name = "    with space before and after  "
    diagram = element_factory.lselect()[0]
    diagram.name = difficult_name
    data = saver()
    loader(data)
    elements = element_factory.lselect()
    assert len(elements) == 2, elements
    assert elements[0].name == difficult_name, elements[0].name


def test_save_and_load_model_with_relationships(
    create, diagram, element_factory, saver, loader
):
    element_factory.create(UML.Package)
    create(CommentItem, UML.Comment)
    create(ClassItem, UML.Class)

    a = diagram.create(AssociationItem)
    a.handles()[0].pos = (10, 20)
    a.handles()[1].pos = (50, 60)
    assert a.handles()[0].pos.x == 10, a.handles()[0].pos
    assert a.handles()[0].pos.y == 20, a.handles()[0].pos
    assert tuple(a.handles()[1].pos) == (50, 60), a.handles()[1].pos

    data = saver()
    loader(data)

    assert len(element_factory.lselect()) == 8
    assert len(element_factory.lselect(UML.Package)) == 1
    assert len(element_factory.lselect(Diagram)) == 1
    d = element_factory.lselect(Diagram)[0]
    assert len(element_factory.lselect(UML.Comment)) == 1
    assert len(element_factory.lselect(UML.Class)) == 1
    assert len(element_factory.lselect(UML.Association)) == 0

    # Check load/save of other canvas items.
    assert len(list(d.get_all_items())) == 3
    aa = next(item for item in d.get_all_items() if isinstance(item, AssociationItem))
    assert aa
    assert list(map(float, aa.handles()[0].pos)) == [10, 20], aa.handles()[0].pos
    assert list(map(float, aa.handles()[1].pos)) == [50, 60], aa.handles()[1].pos
    d1 = next(d.select(lambda e: isinstance(e, ClassItem)))
    assert d1


def test_save_and_load_of_association_with_two_connected_classes(
    create, diagram, element_factory, modeling_language
):
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)
    c2.matrix.translate(200, 200)
    diagram.request_update(c2)
    diagram.update({c1, c2})
    assert tuple(c2.matrix_i2c) == (1, 0, 0, 1, 200, 200)

    a = create(AssociationItem)

    connect(a, a.head, c1)
    connect(a, a.tail, c2)

    diagram.update({c1, c2, a})

    assert a.head.pos.y == 0, a.head.pos
    assert a.tail.pos.x == 200, a.tail.pos
    assert a.tail.pos.y == 200, a.tail.pos
    assert a.subject

    fd = StringIO()
    storage.save(fd, element_factory=element_factory)
    data = fd.getvalue()
    fd.close()

    old_a_subject_id = a.subject.id

    element_factory.flush()
    assert not list(element_factory.select())
    fd = StringIO(data)
    storage.load(
        fd, element_factory=element_factory, modeling_language=modeling_language
    )
    fd.close()

    diagrams = element_factory.lselect(Diagram)
    assert len(diagrams) == 1
    d = diagrams[0]
    a = next(d.select(lambda e: isinstance(e, AssociationItem)))
    assert a.subject is not None
    assert old_a_subject_id == a.subject.id
    cinfo_head = a.diagram.connections.get_connection(a.head)
    assert cinfo_head.connected is not None
    cinfo_tail = a.diagram.connections.get_connection(a.tail)
    assert cinfo_tail.connected is not None
    assert cinfo_head.connected is not cinfo_tail.connected


def test_load_and_save_of_a_model(element_factory, modeling_language, test_models):
    path = test_models / "simple-items.gaphor"

    with open(path, encoding="utf-8") as ifile:
        storage.load(
            ifile,
            element_factory=element_factory,
            modeling_language=modeling_language,
        )

    pf = PseudoFile()

    storage.save(pf, element_factory=element_factory)

    orig = path.read_text(encoding="utf-8")
    copy = pf.data

    expr = re.compile('gaphor-version="[^"]*"')
    orig = expr.sub("%VER%", orig)
    copy = expr.sub("%VER%", copy)

    assert copy == orig, "Saved model does not match copy"


def test_can_not_load_models_older_that_0_17_0(
    element_factory, modeling_language, test_models
):
    path = test_models / "old-gaphor-version.gaphor"

    def load_old_model():
        with open(path, encoding="utf-8") as ifile:
            storage.load(
                ifile,
                element_factory=element_factory,
                modeling_language=modeling_language,
            )

    with pytest.raises(ValueError):
        load_old_model()


def test_save_with_invalid_collection(element_factory, saver, caplog):
    c = element_factory.create(UML.Class)

    p = UML.Package()
    c.package = p
    data = saver()

    assert c.id in data
    assert p.id not in data
    assert "Model has unknown reference" in caplog.text


def test_save_with_invalidreference(element_factory, saver, caplog):
    p = element_factory.create(UML.Package)

    c = UML.Class()
    c.package = p
    data = saver()

    assert p.id in data
    assert c.id not in data
    assert "Model has unknown reference" in caplog.text


def test_save_and_load_with_invalid_element_type(element_factory, saver, loader):
    element_factory.create(UML.Package)

    data = saver()
    data = data.replace("Package", "Foobar")

    with pytest.raises(storage.UnknownModelElementError):
        loader(data)


def test_save_and_load_with_invalid_attribute(element_factory, saver, loader):
    p = element_factory.create(UML.Package)
    p.name = "name"

    data = saver()
    data = data.replace("name", "foobar")
    loader(data)

    package = next(element_factory.select(UML.Package))

    assert not hasattr(package, "foobar")
    assert not package.name


def test_save_and_load_with_invalid_reference(element_factory, saver, loader):
    p1 = element_factory.create(UML.Package)
    p2 = element_factory.create(UML.Package)
    p2.nestingPackage = p1

    data = saver()
    data = data.replace(p1.id, "foobar", 1)
    loader(data)

    package = next(element_factory.select(UML.Package))

    assert not hasattr(package, "foobar")
    assert not package.name


def test_save_and_load_with_enumeration(element_factory, saver, loader):
    p1 = element_factory.create(UML.Package)
    p1.visibility = UML.VisibilityKind.private

    data = saver()
    loader(data)

    package = next(element_factory.select(UML.Package))

    assert package.visibility == UML.VisibilityKind.private



"""Test extend item connections."""

from gaphor import UML
from gaphor.diagram.tests.fixtures import allow, connect, disconnect, get_connected
from gaphor.UML.usecases.extend import ExtendItem
from gaphor.UML.usecases.usecase import UseCaseItem


def test_use_case_glue(create):
    """Test "extend" gluing to use cases."""

    uc1 = create(UseCaseItem, UML.UseCase)
    extend = create(ExtendItem)

    glued = allow(extend, extend.head, uc1)
    assert glued


def test_use_case_connect(create):
    """Test connecting "extend" to use cases."""
    uc1 = create(UseCaseItem, UML.UseCase)
    uc2 = create(UseCaseItem, UML.UseCase)
    extend = create(ExtendItem)

    connect(extend, extend.head, uc1)
    assert get_connected(extend, extend.head), uc1

    connect(extend, extend.tail, uc2)
    assert get_connected(extend, extend.tail), uc2


def test_use_case_reconnect(create):
    """Test reconnecting use cases with "extend"."""
    uc1 = create(UseCaseItem, UML.UseCase)
    uc2 = create(UseCaseItem, UML.UseCase)
    uc3 = create(UseCaseItem, UML.UseCase)
    extend = create(ExtendItem)

    # connect: uc1 -> uc2
    connect(extend, extend.head, uc1)
    connect(extend, extend.tail, uc2)
    e = extend.subject

    # reconnect: uc1 -> uc2
    connect(extend, extend.tail, uc3)

    assert e is not extend.subject
    assert extend.subject.extendedCase is uc1.subject
    assert extend.subject.extension is uc3.subject


def test_use_case_disconnect(create):
    """Test disconnecting "extend" from use cases."""
    uc1 = create(UseCaseItem, UML.UseCase)
    uc2 = create(UseCaseItem, UML.UseCase)
    extend = create(ExtendItem)

    connect(extend, extend.head, uc1)
    connect(extend, extend.tail, uc2)

    disconnect(extend, extend.head)
    assert get_connected(extend, extend.head) is None
    assert extend.subject is None

    disconnect(extend, extend.tail)
    assert get_connected(extend, extend.tail) is None



# ------------------------------------------------------------------
# Copyright (c) 2021 PyInstaller Development Team.
#
# This file is distributed under the terms of the GNU General Public
# License (version 2.0 or later).
#
# The full license is available in LICENSE.GPL.txt, distributed with
# this software.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ------------------------------------------------------------------
# Adapted for Gaphor and Pydot by Arjan Molenaar

import glob
import os
import shutil

from PyInstaller.compat import is_darwin, is_win
from PyInstaller.depend.bindepend import findLibrary

binaries: list[tuple[str, str]] = []
datas: list[tuple[str, str]] = []

# List of binaries agraph.py may invoke.
progs = [
    "dot",
]


def required_plugin(binary):
    return (
        "gvplugin_core" in binary
        or "gvplugin_dot_layout" in binary
        or "gvplugin_neato_layout" in binary
    )


if is_win:
    # For Windows, we move the minimal number of graphviz binaries to a separate folder.
    # This way the GV binaries do not interfere with Gaphor.
    files = [
        "dot.exe",
        "cdt.dll",
        "cgraph++.dll",
        "cgraph.dll",
        "getopt.dll",
        "gvc++.dll",
        "gvc.dll",
        "gvplugin_core.dll",
        "gvplugin_dot_layout.dll",
        "gvplugin_neato_layout.dll",
        "pathplan.dll",
        "vcruntime140.dll",
        "vcruntime140_1.dll",
    ]

    binaries.extend((f"C:/Program Files/Graphviz/bin/{f}", "graphviz") for f in files)

    # Instead of copying the config file, what we should really do
    # is call `dot -c` on our new plugins folder.
    datas.extend(
        (data, "graphviz")
        for data in glob.glob("C:/Program Files/Graphviz/bin/config*")
    )

else:
    if is_darwin:
        # The dot binary in PATH is typically a symlink, handle that.
        # graphviz_bindir is e.g. /usr/local/Cellar/graphviz/2.46.0/bin
        graphviz_bindir = os.path.dirname(os.path.realpath(shutil.which("dot")))  # type: ignore[type-var]

        suffix = "dylib"
        # graphviz_libdir is e.g. /usr/local/Cellar/graphviz/2.46.0/lib/graphviz
        graphviz_libdir = os.path.realpath(f"{graphviz_bindir}/../lib/graphviz")
    else:
        # Do not resolve symlinks: on Ubuntu Bionic, the symlink is
        # /usr/bin/dot -> ../sbin/libgvc6-config-update
        graphviz_bindir = os.path.dirname(shutil.which("dot"))  # type: ignore[type-var]

        suffix = "so.?"
        # graphviz_libdir is e.g. /usr/lib64/graphviz
        graphviz_libdir = os.path.join(
            os.path.dirname(findLibrary("libcdt")), "graphviz"
        )

    binaries.extend((f"{graphviz_bindir}/{binary}", ".") for binary in progs)
    binaries.extend(
        (binary, "graphviz")
        for binary in glob.glob(f"{graphviz_libdir}/*.{suffix}")
        if required_plugin(binary)
    )

    datas.extend((data, "graphviz") for data in glob.glob(f"{graphviz_libdir}/config*"))



from gaphor import UML
from gaphor.diagram.drop import drop
from gaphor.SysML import sysml
from gaphor.SysML.blocks import PropertyItem, ProxyPortItem
from gaphor.UML.recipes import create_connector


def test_drop_sysml_diagram(diagram, element_factory):
    sysml_diagram = element_factory.create(sysml.SysMLDiagram)

    item = drop(sysml_diagram, diagram, 0, 0)

    assert item


def test_drop_connector(diagram, element_factory):
    proxy_port = element_factory.create(sysml.ProxyPort)
    proxy_port_item = diagram.create(ProxyPortItem, subject=proxy_port)
    property = element_factory.create(UML.Property)
    property_item = diagram.create(PropertyItem, subject=property)
    connector = create_connector(proxy_port, property)

    item = drop(connector, diagram, 0, 0)

    assert item
    assert diagram.connections.get_connection(item.head).connected is proxy_port_item
    assert diagram.connections.get_connection(item.tail).connected is property_item


def test_drop_proxy_port(diagram, element_factory):
    block = element_factory.create(sysml.Block)
    proxy_port = element_factory.create(sysml.ProxyPort)
    proxy_port.encapsulatedClassifier = block

    drop(block, diagram, 0, 0)

    item = drop(proxy_port, diagram, 0, 0)

    assert item
    assert (
        diagram.connections.get_connection(item.handles()[0]).connected
        is block.presentation[0]
    )


def test_drop_proxy_port_on_property(diagram, element_factory):
    block = element_factory.create(sysml.Block)
    prop = element_factory.create(UML.Property)
    prop.type = block

    property_item = drop(prop, diagram, 0, 0)
    assert property_item

    port = element_factory.create(sysml.ProxyPort)
    assert not drop(port, diagram, 0, 0), (
        "unable to add port when there is no item that it can be attached to"
    )

    block.ownedPort.append(port)

    port_item = drop(port, diagram, 0, 0)
    assert isinstance(port_item, ProxyPortItem), "port was added"
    assert port_item in property_item.children, "port is attached to the property item"

    property_item_2 = drop(prop, diagram, 500, 500)
    assert property_item_2

    port_item = drop(port, diagram, 500, 500)
    assert isinstance(port_item, ProxyPortItem), "port was added"
    assert port_item in property_item_2.children, (
        "port is attached to the second property item"
    )

    port_item = drop(port, diagram, 100, 200)
    assert isinstance(port_item, ProxyPortItem), "port was added"
    assert port_item in property_item.children, (
        "port is attached to the first property item"
    )



import pytest

from gaphor import UML
from gaphor.diagram.diagramtoolbox import new_item_factory
from gaphor.diagram.tests.fixtures import copy_and_paste_link, copy_clear_and_paste_link
from gaphor.UML.actions.actionstoolbox import partition_config
from gaphor.UML.actions.partition import PartitionItem


@pytest.fixture
def partition_item_factory():
    return new_item_factory(
        PartitionItem,
        UML.ActivityPartition,
        config_func=partition_config,
    )


@pytest.fixture
def partition_item(partition_item_factory, diagram):
    return partition_item_factory(diagram)


def test_partition_placement_adds_two_partitions(partition_item: PartitionItem):
    assert partition_item.subject in partition_item.partition
    assert len(partition_item.partition) == 2


def test_copy_and_paste_partition_item(partition_item, diagram, element_factory):
    new_items = copy_and_paste_link({partition_item}, diagram, element_factory)

    (new_item,) = new_items

    assert new_item.subject is partition_item.subject
    assert new_item.partition[0] is partition_item.partition[0]
    assert new_item.partition[1] is partition_item.partition[1]


def test_copy_remove_and_paste_partition_item(partition_item, diagram, element_factory):
    copy_clear_and_paste_link({partition_item}, diagram, element_factory)

    new_item = diagram.ownedPresentation[0]

    assert new_item.subject in new_item.partition
    assert len(new_item.partition) == 2



from gaphor import UML
from gaphor.core.modeling import Diagram, ElementFactory
from gaphor.diagram.copypaste import copy_full, paste_full
from gaphor.diagram.tests.test_copypaste_link import two_classes_and_a_generalization
from gaphor.UML.classes import ClassItem, GeneralizationItem, PackageItem
from gaphor.UML.general import DiagramItem


def test_copied_item_references_new_model_element(diagram, element_factory):
    cls = element_factory.create(UML.Class)
    cls.name = "Name"
    cls_item = diagram.create(ClassItem, subject=cls)

    buffer = copy_full({cls_item})

    all(paste_full(buffer, diagram))

    assert len(list(diagram.get_all_items())) == 2
    item1, item2 = diagram.get_all_items()

    assert item1.subject
    assert item2.subject
    assert item1.subject in element_factory
    assert item2.subject in element_factory
    assert item1.subject is not item2.subject
    assert item1.subject.name == item2.subject.name


def test_copy_multiple_items(diagram, element_factory):
    cls = element_factory.create(UML.Class)
    cls_item1 = diagram.create(ClassItem, subject=cls)
    cls_item2 = diagram.create(ClassItem, subject=cls)

    buffer = copy_full({cls_item1, cls_item2})

    paste_full(buffer, diagram)

    assert len(list(diagram.get_all_items())) == 4
    assert len(element_factory.lselect(UML.Class)) == 2


def test_copy_items_with_connections(diagram, element_factory):
    spc_cls_item, gen_cls_item, gen_item = two_classes_and_a_generalization(
        diagram, element_factory
    )

    buffer = copy_full({gen_cls_item, gen_item, spc_cls_item})
    new_items = paste_full(buffer, diagram)
    (new_cls1, new_cls2) = (ci.subject for ci in new_items if isinstance(ci, ClassItem))
    (new_gen,) = (gi.subject for gi in new_items if isinstance(gi, GeneralizationItem))

    assert new_gen.general in {new_cls1, new_cls2}
    assert new_gen.specific in {new_cls1, new_cls2}


def test_copy_element_factory(diagram, element_factory):
    """This case tests the boundaries of the copy/paste functionality:

    it copies a complete element factory, something that's not possible
    to do from the UI.
    """
    two_classes_and_a_generalization(diagram, element_factory)

    buffer = copy_full(element_factory)

    new_element_factory = ElementFactory()
    new_diagram = new_element_factory.create(UML.Diagram)
    paste_full(buffer, new_diagram)

    assert len(new_element_factory.lselect(Diagram)) == 1
    assert element_factory.size() == new_element_factory.size()


def test_copy_package_with_diagram(element_factory):
    diagram: UML.Diagram = element_factory.create(UML.Diagram)
    package = element_factory.create(UML.Package)
    package.ownedDiagram = diagram
    package_item = diagram.create(PackageItem, subject=package)

    copy_buffer = copy_full([package_item])
    (new_package_item,) = paste_full(copy_buffer, diagram)
    new_package = new_package_item.subject

    assert diagram in package.ownedDiagram
    assert diagram not in new_package.ownedDiagram
    assert diagram.element is package


def test_shallow_copy_package_with_owned_package(element_factory):
    diagram: UML.Diagram = element_factory.create(UML.Diagram)
    package = element_factory.create(UML.Package)
    subpackage = element_factory.create(UML.Package)
    subpackage.nestingPackage = package
    package_item = diagram.create(PackageItem, subject=package)

    copy_buffer = copy_full([package_item])
    (new_package_item,) = paste_full(copy_buffer, diagram)
    new_package = new_package_item.subject

    assert subpackage in package.nestedPackage
    assert subpackage.nestingPackage is package
    assert subpackage not in new_package.nestedPackage


def test_full_copy_package_with_owned_package(element_factory):
    diagram: UML.Diagram = element_factory.create(UML.Diagram)
    package = element_factory.create(UML.Package)
    subpackage = element_factory.create(UML.Package)
    subpackage.nestingPackage = package
    subpackage.name = "subpackage"
    package_item = diagram.create(PackageItem, subject=package)

    copy_buffer = copy_full([package_item], element_factory.lookup)
    (new_package_item,) = paste_full(copy_buffer, diagram)
    new_package = new_package_item.subject

    assert subpackage in package.nestedPackage
    assert subpackage.nestingPackage is package
    assert new_package.nestedPackage
    assert subpackage not in new_package.nestedPackage
    assert new_package.nestedPackage[0].name == "subpackage"


def test_copy_diagram_with_elements(diagram, element_factory):
    other_diagram: UML.Diagram = element_factory.create(UML.Diagram)
    two_classes_and_a_generalization(other_diagram, element_factory)

    diagram_item = diagram.create(DiagramItem, subject=other_diagram)

    copy_buffer = copy_full([diagram_item], element_factory.lookup)

    target_diagram: UML.Diagram = element_factory.create(UML.Diagram)
    (new_diagram_item, *_other) = paste_full(copy_buffer, target_diagram)

    new_diagram = new_diagram_item.subject

    assert new_diagram.ownedPresentation



"""Test messages."""

import pytest
from gaphas.item import SE

from gaphor import UML
from gaphor.core.modeling import Diagram
from gaphor.diagram.group import group
from gaphor.diagram.tests.fixtures import connect
from gaphor.UML.interactions.interaction import InteractionItem
from gaphor.UML.interactions.lifeline import LifelineItem
from gaphor.UML.interactions.message import MessageItem


def test_message_persistence(diagram, element_factory, saver, loader):
    diagram.create(MessageItem, subject=element_factory.create(UML.Message))

    data = saver()
    loader(data)
    new_diagram = next(element_factory.select(Diagram))
    item = next(new_diagram.select(MessageItem))

    assert item


def test_group_message_item_without_subject(diagram, element_factory):
    interaction = diagram.create(
        InteractionItem, subject=element_factory.create(UML.Interaction)
    )
    message = diagram.create(MessageItem)

    with pytest.raises(TypeError):
        group(interaction.subject, message.subject)

    assert message.subject is None


def test_group_message_item_with_subject(diagram, element_factory):
    interaction = diagram.create(
        InteractionItem, subject=element_factory.create(UML.Interaction)
    )
    message = diagram.create(MessageItem, subject=element_factory.create(UML.Message))

    group(interaction.subject, message.subject)

    assert message.subject
    assert message.subject.interaction is interaction.subject


def test_message_connected_to_lifeline(diagram, element_factory, saver, loader):
    lifeline = diagram.create(
        LifelineItem, subject=element_factory.create(UML.Lifeline)
    )
    lifeline.lifetime.length = 100
    message = diagram.create(MessageItem, subject=element_factory.create(UML.Message))
    diagram.update({lifeline, message})
    message.head.pos.x = lifeline.handles()[SE].pos.x
    message.head.pos.y = lifeline.handles()[-1].pos.y

    connect(message, message.head, lifeline)
    assert lifeline.lifetime.visible
    assert (
        diagram.connections.get_connection(message.head).port is lifeline.lifetime.port
    )

    data = saver()
    loader(data)

    new_diagram = next(element_factory.select(Diagram))
    new_lifeline = next(new_diagram.select(LifelineItem))
    new_message = next(new_diagram.select(MessageItem))
    new_port = new_diagram.connections.get_connection(new_message.head).port

    assert new_lifeline.lifetime.visible
    assert new_lifeline.lifetime.length == 100
    assert new_port is new_lifeline.lifetime.port



import pytest

from gaphor.ui.styling import Styling


class MockProperties(dict):
    def set(self, key, value):
        self[key] = value


@pytest.fixture
def properties():
    return MockProperties()


@pytest.fixture(scope="session", autouse=True)
def styling():
    styling = Styling()
    yield styling
    styling.shutdown()



import subprocess

import pytest
from dulwich import porcelain
from dulwich.repo import Repo


def create_merge_conflict(repo: Repo, filename, initial_text, our_text, their_text):
    def commit_all(message):
        porcelain.add(repo, paths=[filename])
        return porcelain.commit(
            repo, author=b"Alice Author <alice@gaphor.org>", message=message
        )

    filename.write_text(initial_text, encoding="utf-8")
    main_ref = commit_all("Initial commit")

    porcelain.branch_create(repo, "branch")
    porcelain.switch(repo, "branch")
    filename.write_text(their_text, encoding="utf-8")
    branch_oid = commit_all("Branch commit")

    porcelain.switch(repo, main_ref, detach=True, force=True)
    filename.write_text(our_text, encoding="utf-8")
    commit_all("Second commit")

    # `porcelain.merge` function does not exist.
    _porcelain_merge(repo, branch_oid)


def _porcelain_merge(repo, branch):
    """Merge branch with git cli."""

    subprocess.run(
        ["git", "config", "user.name", "Alice Author"], cwd=repo.path, check=True
    )
    subprocess.run(
        ["git", "config", "user.email", "alice@gaphor.org"], cwd=repo.path, check=True
    )
    with pytest.raises(subprocess.CalledProcessError):
        subprocess.run(["git", "merge", branch], cwd=repo.path, check=True)



from gaphor.core.changeset.compare import set_value_change_property_value
from gaphor.core.modeling import ElementChange, ValueChange
from gaphor.ui.modelmerge.editor import ModelMerge


def test_build_list_store(event_manager, element_factory, modeling_language):
    model_merge = ModelMerge(event_manager, element_factory, modeling_language)

    change: ElementChange = element_factory.create(ElementChange)
    change.op = "add"
    change.element_id = "1234"
    change.element_name = "Diagram"
    vchange: ValueChange = element_factory.create(ValueChange)
    vchange.op = "update"
    vchange.element_id = "1234"
    vchange.property_name = "name"
    set_value_change_property_value(vchange, "my diagram")

    model_merge.refresh_model()

    assert model_merge.model
    assert not model_merge.model[0].children



"""Test handle tool functionality."""

import pytest
from gaphas.connector import ConnectionSink
from gaphas.connector import Connector as ConnectorAspect
from gaphas.handlemove import HandleMove
from gi.repository import GLib, Gtk

from gaphor import UML
from gaphor.core.modeling import Diagram
from gaphor.diagram._connector import PresentationConnector
from gaphor.diagram.connectors import Connector
from gaphor.diagram.event import DiagramOpened
from gaphor.ui.diagrams import Diagrams
from gaphor.ui.toolbox import Toolbox
from gaphor.UML.general import CommentItem, CommentLineItem
from gaphor.UML.usecases.actor import ActorItem


@pytest.fixture
def diagrams(event_manager, element_factory, modeling_language, properties):
    toolbox = Toolbox(event_manager, properties, modeling_language)
    diagrams = Diagrams(
        event_manager, element_factory, properties, modeling_language, toolbox
    )
    window = Gtk.Window.new()
    box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
    window.set_child(box)
    box.append(diagrams.open())
    window.set_visible(True)
    yield diagrams
    diagrams.close()


@pytest.fixture
def connections(diagram):
    return diagram.connections


@pytest.fixture
def comment(element_factory, diagram):
    return diagram.create(CommentItem, subject=element_factory.create(UML.Comment))


@pytest.fixture
def commentline(diagram):
    return diagram.create(CommentLineItem)


def test_aspect_type(commentline, connections):
    aspect = ConnectorAspect(commentline, commentline.handles()[0], connections)
    assert isinstance(aspect, PresentationConnector)


def test_query(comment, commentline):
    assert Connector(comment, commentline)


def test_allow(commentline, comment, connections):
    aspect = ConnectorAspect(commentline, commentline.handles()[0], connections)
    assert aspect.item is commentline
    assert aspect.handle is commentline.handles()[0]

    sink = ConnectionSink(comment)
    assert aspect.allow(sink)


def test_connect(diagram, comment, commentline, connections):
    sink = ConnectionSink(comment)
    aspect = ConnectorAspect(commentline, commentline.handles()[0], connections)
    aspect.connect(sink)
    cinfo = diagram.connections.get_connection(commentline.handles()[0])
    assert cinfo, cinfo


def current_diagram_view(diagrams):
    """Get a view for the current diagram."""
    view = diagrams.get_current_view()

    # realize view, forces bounding box recalculation
    ctx = GLib.main_context_default()
    while ctx.pending():
        ctx.iteration(False)

    return view


@pytest.mark.asyncio
async def test_iconnect(event_manager, element_factory, diagrams):
    """Test basic glue functionality using CommentItem and CommentLine
    items."""
    diagram = element_factory.create(Diagram)
    event_manager.handle(DiagramOpened(diagram))
    comment = diagram.create(CommentItem, subject=element_factory.create(UML.Comment))

    line = diagram.create(CommentLineItem)

    view = current_diagram_view(diagrams)
    assert view, "View should be available here"

    # select handle:
    handle = line.handles()[-1]

    move = HandleMove(line, handle, view)
    handle.pos = (0, 0)
    await view.update()
    item = move.glue(handle.pos)
    assert item is not None

    move.connect(handle.pos)
    cinfo = diagram.connections.get_connection(handle)
    assert cinfo.constraint is not None
    assert cinfo.connected is comment, cinfo.connected

    ConnectorAspect(line, handle, diagram.connections).disconnect()

    cinfo = diagram.connections.get_connection(handle)

    assert cinfo is None


@pytest.mark.asyncio
async def test_connect_comment_and_actor(event_manager, element_factory, diagrams):
    """Test connect/disconnect on comment and actor using comment-line."""
    diagram = element_factory.create(Diagram)
    event_manager.handle(DiagramOpened(diagram))
    comment = diagram.create(CommentItem, subject=element_factory.create(UML.Comment))

    line = diagram.create(CommentLineItem)
    view = current_diagram_view(diagrams)
    assert view, "View should be available here"

    handle = line.handles()[0]
    move = HandleMove(line, handle, view)

    handle.pos = (0, 0)
    await view.update()
    sink = move.glue(handle.pos)

    assert sink is not None
    assert sink.item is comment

    move.connect(handle.pos)
    cinfo = diagram.connections.get_connection(handle)
    assert cinfo is not None, None
    assert cinfo.item is line
    assert cinfo.connected is comment

    # Connect the other end to the Actor:
    actor = diagram.create(ActorItem, subject=element_factory.create(UML.Actor))

    handle = line.handles()[-1]
    move = HandleMove(line, handle, view)

    handle.pos = (0, 0)
    await view.update()
    sink = move.glue(handle.pos)

    assert sink, f"No sink at {handle.pos}"
    assert sink.item is actor

    move.connect(handle.pos)
    cinfo = view.model.connections.get_connection(handle)

    assert cinfo.item is line
    assert cinfo.connected is actor

    # Try to connect far away from any item will only do a full disconnect
    assert len(comment.subject.annotatedElement) == 1, comment.subject.annotatedElement
    assert actor.subject in comment.subject.annotatedElement

    sink = move.glue((500, 500))
    assert sink is None, sink
    move.connect((500, 500))

    cinfo = view.model.connections.get_connection(handle)
    assert cinfo is None



import pytest
from gaphas.segment import Segment

from gaphor import UML
from gaphor.core.modeling import (
    Diagram,
    ElementFactory,
    StyleSheet,
    swap_element_type,
)
from gaphor.diagram.general import Line
from gaphor.diagram.tests.fixtures import connect, disconnect
from gaphor.storage.recovery import Recorder, replay_events
from gaphor.UML.diagramitems import ClassItem, DependencyItem
from gaphor.UML.general import CommentItem


@pytest.fixture
def recorder(event_manager):
    recorder = Recorder()
    recorder.subscribe(event_manager)
    yield recorder
    recorder.unsubscribe(event_manager)


def test_record_create_element(
    recorder, event_manager, element_factory, modeling_language
):
    comment = element_factory.create(UML.Comment)

    new_model = ElementFactory(event_manager)

    replay_events(recorder.events[:], new_model, modeling_language)

    assert ("c", "UML", "Comment", comment.id, None) in recorder.events
    assert new_model.lookup(comment.id)


def test_record_create_presentation(
    recorder, event_manager, diagram, modeling_language
):
    comment = diagram.create(CommentItem)

    new_model = ElementFactory(event_manager)
    new_model.create_as(UML.Diagram, id=diagram.id)

    replay_events(recorder.events[:], new_model, modeling_language)

    assert ("c", "UML", "CommentItem", comment.id, diagram.id) in recorder.events

    assert new_model.lookup(comment.id)


def test_record_delete_element(
    recorder, event_manager, element_factory, modeling_language
):
    comment = element_factory.create(UML.Comment)
    comment.unlink()

    new_model = ElementFactory(event_manager)
    replay_events(recorder.events[:], new_model, modeling_language)

    assert ("c", "UML", "Comment", comment.id, None) in recorder.events
    assert ("u", comment.id, None) in recorder.events
    assert not new_model.lookup(comment.id)


def test_record_delete_presentation(
    recorder, event_manager, diagram, modeling_language
):
    comment = diagram.create(CommentItem)
    comment.unlink()

    new_model = ElementFactory(event_manager)
    new_model.create_as(UML.Diagram, diagram.id)
    replay_events(recorder.events[:], new_model, modeling_language)

    assert ("c", "UML", "CommentItem", comment.id, diagram.id) in recorder.events
    assert ("u", comment.id, diagram.id) in recorder.events
    assert not new_model.lookup(comment.id)


def test_record_update_attribute(
    recorder, event_manager, element_factory, modeling_language
):
    comment = element_factory.create(UML.Comment)
    comment.body = "Foo"

    new_model = ElementFactory(event_manager)
    replay_events(recorder.events[:], new_model, modeling_language)

    assert ("c", "UML", "Comment", comment.id, None) in recorder.events
    assert ("a", comment.id, "body", "Foo") in recorder.events
    assert new_model.lookup(comment.id)
    assert new_model.lookup(comment.id).body == "Foo"


def test_record_set_association(
    recorder, event_manager, element_factory, modeling_language
):
    comment = element_factory.create(UML.Comment)
    diagram = element_factory.create(UML.Diagram)
    diagram.element = comment

    new_model = ElementFactory(event_manager)
    replay_events(recorder.events[:], new_model, modeling_language)

    new_comment = new_model.lookup(comment.id)
    new_diagram = new_model.lookup(diagram.id)

    assert ("s", diagram.id, "element", comment.id) in recorder.events
    assert ("s", comment.id, "ownedDiagram", diagram.id) in recorder.events

    assert new_comment is new_diagram.element
    assert new_diagram in new_comment.ownedDiagram


def test_record_delete_association(
    recorder, event_manager, element_factory, modeling_language
):
    comment = element_factory.create(UML.Comment)
    diagram = element_factory.create(UML.Diagram)
    diagram.element = comment
    del diagram.element

    new_model = ElementFactory(event_manager)
    replay_events(recorder.events[:], new_model, modeling_language)

    new_comment = new_model.lookup(comment.id)
    new_diagram = new_model.lookup(diagram.id)

    assert ("s", diagram.id, "element", None) in recorder.events
    assert ("d", comment.id, "ownedDiagram", diagram.id) in recorder.events

    assert not new_diagram.element
    assert new_diagram not in new_comment.ownedDiagram


def test_record_move_element(
    recorder, event_manager, element_factory, modeling_language
):
    diagram = element_factory.create(Diagram)
    class_item = diagram.create(ClassItem, subject=element_factory.create(UML.Class))
    class_item.matrix.translate(200, 100)

    new_model = ElementFactory(event_manager)
    replay_events(recorder.events[:], new_model, modeling_language)

    new_class_item = new_model.lookup(class_item.id)

    assert new_class_item.matrix[4] == pytest.approx(200)
    assert new_class_item.matrix[5] == pytest.approx(100)


def test_record_move_handle_element(
    recorder, event_manager, element_factory, modeling_language
):
    diagram = element_factory.create(Diagram)
    class_item = diagram.create(ClassItem, subject=element_factory.create(UML.Class))
    class_item.handles()[2].pos = (200, 100)

    new_model = ElementFactory(event_manager)
    replay_events(recorder.events[:], new_model, modeling_language)

    new_class_item = new_model.lookup(class_item.id)

    assert new_class_item.handles()[2].pos.x == 200
    assert new_class_item.handles()[2].pos.y == 100


def test_record_connect_element(
    recorder, event_manager, element_factory, modeling_language
):
    diagram = element_factory.create(Diagram)
    class_item = diagram.create(ClassItem, subject=element_factory.create(UML.Class))
    dependency_item = diagram.create(DependencyItem)
    connect(dependency_item, dependency_item.head, class_item)

    new_model = ElementFactory(event_manager)
    replay_events(recorder.events[:], new_model, modeling_language)

    new_diagram = new_model.lookup(diagram.id)
    new_class_item = new_model.lookup(class_item.id)
    new_dependency_item = new_model.lookup(dependency_item.id)
    cinfo = new_diagram.connections.get_connection(new_dependency_item.head)

    assert cinfo
    assert cinfo.item is new_dependency_item
    assert cinfo.connected is new_class_item


def test_record_connect_both_ends(
    recorder, event_manager, element_factory, modeling_language
):
    diagram = element_factory.create(UML.Diagram)
    class_item = diagram.create(ClassItem, subject=element_factory.create(UML.Class))
    other_class_item = diagram.create(
        ClassItem, subject=element_factory.create(UML.Class)
    )
    dependency_item = diagram.create(DependencyItem)
    connect(dependency_item, dependency_item.head, class_item)
    connect(dependency_item, dependency_item.tail, other_class_item)

    new_model = ElementFactory(event_manager)
    replay_events(recorder.events[:], new_model, modeling_language)

    new_diagram = new_model.lookup(diagram.id)
    new_class_item = new_model.lookup(class_item.id)
    new_other_class_item = new_model.lookup(other_class_item.id)
    new_dependency_item = new_model.lookup(dependency_item.id)
    cinfo_head = new_diagram.connections.get_connection(new_dependency_item.head)
    cinfo_tail = new_diagram.connections.get_connection(new_dependency_item.tail)

    assert new_dependency_item.subject
    assert new_class_item.subject in new_dependency_item.subject.supplier
    assert new_other_class_item.subject in new_dependency_item.subject.client
    assert cinfo_head
    assert cinfo_tail
    assert cinfo_head.item is new_dependency_item
    assert cinfo_head.connected is new_class_item


def test_record_disconnect_element(
    recorder, event_manager, element_factory, modeling_language
):
    diagram = element_factory.create(UML.Diagram)
    class_item = diagram.create(ClassItem, subject=element_factory.create(UML.Class))
    other_class_item = diagram.create(
        ClassItem, subject=element_factory.create(UML.Class)
    )
    dependency_item = diagram.create(DependencyItem)
    connect(dependency_item, dependency_item.head, class_item)
    connect(dependency_item, dependency_item.tail, other_class_item)
    disconnect(dependency_item, dependency_item.tail)

    new_model = ElementFactory(event_manager)
    replay_events(recorder.events[:], new_model, modeling_language)

    new_diagram = new_model.lookup(diagram.id)
    new_class_item = new_model.lookup(class_item.id)
    new_dependency_item = new_model.lookup(dependency_item.id)
    cinfo_head = new_diagram.connections.get_connection(new_dependency_item.head)
    cinfo_tail = new_diagram.connections.get_connection(new_dependency_item.tail)

    assert cinfo_head
    assert cinfo_head.item is new_dependency_item
    assert cinfo_head.connected is new_class_item
    assert not cinfo_tail


def test_record_reconnect_element(
    recorder, event_manager, element_factory, modeling_language
):
    diagram = element_factory.create(Diagram)
    class_item = diagram.create(ClassItem, subject=element_factory.create(UML.Class))
    dependency_item = diagram.create(DependencyItem)
    connect(dependency_item, dependency_item.head, class_item)
    dependency_item.head.pos = (100, 100)
    connect(dependency_item, dependency_item.head, class_item)
    diagram.update()

    new_model = ElementFactory(event_manager)
    replay_events(recorder.events[:], new_model, modeling_language)

    new_diagram = new_model.lookup(diagram.id)
    new_dependency_item = new_model.lookup(dependency_item.id)
    new_diagram.update()

    assert ("ir", dependency_item.id, 0, class_item.id, 1) in recorder.events
    assert dependency_item.head.pos == new_dependency_item.head.pos


def test_record_split_line_segments(
    recorder, event_manager, element_factory, modeling_language
):
    diagram = element_factory.create(Diagram)
    line_item = diagram.create(Line)
    Segment(line_item, diagram).split_segment(0, 2)
    handle_positions = [tuple(h.pos) for h in line_item.handles()]

    new_model = ElementFactory(event_manager)
    replay_events(recorder.events[:], new_model, modeling_language)

    new_line_item = new_model.lookup(line_item.id)
    new_handle_positions = [tuple(h.pos) for h in new_line_item.handles()]

    assert len(new_handle_positions) == 3
    assert handle_positions == new_handle_positions


def test_record_merge_line_segments(
    recorder, event_manager, element_factory, modeling_language
):
    diagram = element_factory.create(Diagram)
    line_item = diagram.create(Line)
    Segment(line_item, diagram).split_segment(0, 3)
    new_model = ElementFactory(event_manager)
    replay_events(recorder.events[:], new_model, modeling_language)

    recorder.truncate()
    Segment(line_item, diagram).merge_segment(1, 2)
    handle_positions = [tuple(h.pos) for h in line_item.handles()]
    replay_events(recorder.events[:], new_model, modeling_language)

    new_line_item = new_model.lookup(line_item.id)
    new_handle_positions = [tuple(h.pos) for h in new_line_item.handles()]

    assert len(new_handle_positions) == 3
    assert handle_positions == new_handle_positions


def test_record_type_swapped(
    recorder, event_manager, element_factory, modeling_language
):
    comment = element_factory.create(UML.Comment)
    swap_element_type(comment, StyleSheet)
    new_model = ElementFactory(event_manager)
    replay_events(recorder.events[:], new_model, modeling_language)

    new_diagram = new_model.lookup(comment.id)

    assert new_diagram
    assert isinstance(new_diagram, StyleSheet)
    assert all(isinstance(f, str) for f in recorder.events[-1]), recorder.events[-1]



from gaphor.diagram.group import group, ungroup
from gaphor.SysML.sysml import Block
from gaphor.UML import Property


def test_group(element_factory):
    property = element_factory.create(Property)
    block = element_factory.create(Block)

    assert group(block, property)

    assert property.owner is block


def test_ungroup(element_factory):
    property = element_factory.create(Property)
    block = element_factory.create(Block)
    group(block, property)

    assert ungroup(block, property)

    assert property.owner is None


def test_do_not_ungroup_wrong_parent(element_factory):
    property = element_factory.create(Property)
    block = element_factory.create(Block)
    wrong_block = element_factory.create(Block)
    group(block, property)

    assert not ungroup(wrong_block, property)

    assert property.owner is block



import pytest

from gaphor import UML
from gaphor.diagram.tests.fixtures import find
from gaphor.UML.classes.classespropertypages import (
    AttributesPage,
    ClassifierPropertyPage,
    Folded,
    InterfacePropertyPage,
    NamedElementPropertyPage,
    OperationsPage,
    ShowAttributesPage,
    ShowOperationsPage,
    attribute_model,
)


@pytest.fixture
def metaclass(element_factory):
    klass = element_factory.create(UML.Class)
    stereotype = element_factory.create(UML.Stereotype)
    UML.recipes.create_extension(klass, stereotype)
    return klass


def test_named_element_property_page(element_factory, event_manager):
    subject = element_factory.create(UML.Class)
    property_page = NamedElementPropertyPage(subject, event_manager)

    widget = property_page.construct()
    name_entry = find(widget, "name-entry")
    name_entry.set_text("Name")

    assert subject.name == "Name"


def test_no_named_element_property_page_for_metaclass(metaclass, event_manager):
    property_page = NamedElementPropertyPage(metaclass, event_manager)
    widget = property_page.construct()

    assert widget is None


def test_classifier_property_page(element_factory, event_manager):
    subject = element_factory.create(UML.Class)
    property_page = ClassifierPropertyPage(subject, event_manager)

    widget = property_page.construct()
    abstract = find(widget, "abstract")
    abstract.set_active(True)
    is_final_specialization = find(widget, "is-final-specialization")
    is_final_specialization.set_active(True)

    assert subject.isAbstract
    assert subject.isFinalSpecialization


def test_no_classifier_property_page_for_metaclass(metaclass, event_manager):
    property_page = ClassifierPropertyPage(metaclass, event_manager)
    widget = property_page.construct()

    assert widget is None


def test_interface_property_page(diagram, element_factory, event_manager):
    item = diagram.create(
        UML.classes.InterfaceItem, subject=element_factory.create(UML.Interface)
    )
    property_page = InterfacePropertyPage(item, event_manager)

    widget = property_page.construct()
    folded = find(widget, "folded")
    folded.set_active(True)

    assert item.folded == Folded.PROVIDED


def test_show_attributes_page(diagram, element_factory, event_manager):
    item = diagram.create(
        UML.classes.InterfaceItem, subject=element_factory.create(UML.Interface)
    )
    property_page = ShowAttributesPage(item, event_manager)

    widget = property_page.construct()
    show_attributes = find(widget, "show-attributes")
    show_attributes.set_active(False)

    assert not item.show_attributes


def test_attributes_page_add_attribute(element_factory, event_manager):
    subject = element_factory.create(UML.Interface)

    property_page = AttributesPage(subject, event_manager)

    property_page.construct()
    property_page.model.get_item(0).attribute = "+ attr: str"

    assert subject.attribute[0].name == "attr"
    assert subject.attribute[0].typeValue == "str"


def test_attribute_create_model(element_factory, event_manager):
    subject = element_factory.create(UML.Class)

    attr1 = element_factory.create(UML.Property)
    attr1.name = "attr1"
    subject.ownedAttribute = attr1
    attr2 = element_factory.create(UML.Property)
    attr2.name = "attr2"
    subject.ownedAttribute = attr2
    attr3 = element_factory.create(UML.Property)
    attr3.name = "attr3"
    subject.ownedAttribute = attr3

    list_store = attribute_model(subject, event_manager)

    assert list_store[0].attr is attr1
    assert list_store[1].attr is attr2
    assert list_store[2].attr is attr3


def test_attribute_change_order(element_factory, event_manager):
    subject = element_factory.create(UML.Class)

    attr1 = element_factory.create(UML.Property)
    attr1.name = "attr1"
    subject.ownedAttribute = attr1
    attr2 = element_factory.create(UML.Property)
    attr2.name = "attr2"
    subject.ownedAttribute = attr2
    attr3 = element_factory.create(UML.Property)
    attr3.name = "attr3"
    subject.ownedAttribute = attr3

    property_page = AttributesPage(subject, event_manager)
    widget = property_page.construct()
    list_store = find(widget, "attributes-list").get_model()

    subject.ownedAttribute.swap(attr1, attr2)

    assert list_store[0].attr is attr2
    assert list_store[1].attr is attr1
    assert list_store[2].attr is attr3


def test_show_operations_page(diagram, element_factory, event_manager):
    item = diagram.create(
        UML.classes.InterfaceItem, subject=element_factory.create(UML.Interface)
    )
    property_page = ShowOperationsPage(item, event_manager)

    widget = property_page.construct()
    show_operations = find(widget, "show-operations")
    show_operations.set_active(False)

    assert not item.show_operations


def test_operations_page_add_operation(element_factory, event_manager):
    subject = element_factory.create(UML.Interface)

    property_page = OperationsPage(subject, event_manager)

    property_page.construct()
    property_page.model.get_item(0).operation = "+ oper()"

    assert subject.ownedOperation[0].name == "oper"



from gaphor.ui.installschemas import install_schemas_parser


def test_install_schemas(tmp_path):
    parser = install_schemas_parser()

    args = parser.parse_args(["--schemas-dir", str(tmp_path)])

    args.command(args)

    assert (tmp_path / "org.gaphor.Gaphor.gschema.xml").exists()
    assert (tmp_path / "gschemas.compiled").exists()



from gaphor import UML
from gaphor.diagram.group import group, ungroup


def test_node_group(element_factory):
    """Test node within another node composition."""
    n1 = element_factory.create(UML.Node)
    n2 = element_factory.create(UML.Node)

    assert group(n1, n2)

    assert n2 in n1.nestedNode
    assert n1 not in n2.nestedNode


def test_node_ungroup(element_factory):
    """Test decomposition of component from node."""
    n1 = element_factory.create(UML.Node)
    n2 = element_factory.create(UML.Node)

    assert group(n1, n2)
    assert ungroup(n1, n2)

    assert n2 not in n1.nestedNode
    assert n1 not in n2.nestedNode


def test_node_ungroup_wrong_parent(element_factory):
    """Test decomposition of component from node."""
    n1 = element_factory.create(UML.Node)
    n2 = element_factory.create(UML.Node)
    wrong = element_factory.create(UML.Node)

    assert group(n1, n2)
    assert not ungroup(wrong, n2)

    assert n2 in n1.nestedNode
    assert n1 not in n2.nestedNode


def test_artifact_group(element_factory):
    """Test artifact within node deployment."""
    n = element_factory.create(UML.Node)
    a = element_factory.create(UML.Artifact)

    assert group(n, a)

    assert n.deployment
    assert n.deployment[0].deployedArtifact[0] is a
    assert len(element_factory.lselect(UML.Deployment)) == 1


def test_artifact_group_is_idempotent(element_factory):
    """Test artifact within node deployment."""
    n = element_factory.create(UML.Node)
    a = element_factory.create(UML.Artifact)

    assert group(n, a)
    assert group(n, a)
    assert group(n, a)

    assert n.deployment
    assert n.deployment[0].deployedArtifact[0] is a
    assert len(element_factory.lselect(UML.Deployment)) == 1


def test_artifact_ungroup(element_factory):
    """Test removal of artifact from node."""
    n = element_factory.create(UML.Node)
    a = element_factory.create(UML.Artifact)

    assert group(n, a)
    assert ungroup(n, a)

    assert not n.deployment
    assert not element_factory.lselect(UML.Deployment)


def test_artifact_ungroup_wrong_parent(element_factory):
    """Test removal of artifact from node."""
    n = element_factory.create(UML.Node)
    a = element_factory.create(UML.Artifact)
    wrong = element_factory.create(UML.Node)

    assert group(n, a)
    assert not ungroup(wrong, a)

    assert n.deployment
    assert element_factory.lselect(UML.Deployment)



import textwrap

import pytest

import gaphor.UML.classes.classespropertypages  # noqa
from gaphor import UML
from gaphor.core.modeling.diagram import StyledItem
from gaphor.diagram.tests.fixtures import find
from gaphor.services.componentregistry import ComponentRegistry
from gaphor.ui.elementeditor import ElementEditor, dump_css_tree
from gaphor.UML.diagramitems import ClassItem, PackageItem


class DiagramsStub:
    def get_current_view(self):
        return None


@pytest.fixture
def diagrams():
    return DiagramsStub()


class MockMainWindow:
    @property
    def window(self):
        return None


@pytest.fixture
def component_registry(event_manager, element_factory):
    reg = ComponentRegistry()
    reg.register("event_manager", event_manager)
    reg.register("element_factory", element_factory)
    reg.register("main_window", MockMainWindow())
    return reg


class DummyProperties(dict):
    def set(self, key, val):
        self[key] = val


def test_reopen_of_window(
    event_manager, component_registry, element_factory, modeling_language, diagrams
):
    properties = DummyProperties()
    editor = ElementEditor(
        event_manager,
        component_registry,
        element_factory,
        modeling_language,
        diagrams,
        properties,
    )

    editor.open()
    editor.close()
    editor.open()
    editor.close()


def test_create_pages(
    event_manager,
    component_registry,
    element_factory,
    modeling_language,
    diagrams,
    create,
):
    properties = DummyProperties()
    editor = ElementEditor(
        event_manager,
        component_registry,
        element_factory,
        modeling_language,
        diagrams,
        properties,
    )
    package_item = create(PackageItem, UML.Package)

    editor.open()
    editor.editors.create_pages(package_item)

    assert find(editor.editors.vbox, "name-editor")

    editor.editors.clear_pages()

    assert not find(editor.editors.vbox, "name-editor")


def test_dump_css_tree(element_factory, create):
    class_item = create(ClassItem, UML.Class)

    class_item.subject.ownedAttribute = element_factory.create(UML.Property)
    class_item.subject.ownedOperation = element_factory.create(UML.Operation)

    text = dump_css_tree(StyledItem(class_item))

    assert text == textwrap.dedent(
        """\
        class .item
         ├╴compartment
         │  ├╴stereotypes
         │  ├╴name
         │  ╰╴from
         ├╴compartment
         │  ╰╴attribute
         ╰╴compartment
            ╰╴operation"""
    )



# ruff: noqa: F401,F811
import pytest

from gaphor import UML
from gaphor.conftest import (
    create,
    diagram,
    element_factory,
    event_manager,
    modeling_language,
)
from gaphor.core import Transaction
from gaphor.diagram.tests.fixtures import connect
from gaphor.services.undomanager import UndoManager
from gaphor.UML.classes import AssociationItem, ClassItem
from gaphor.UML.sanitizerservice import SanitizerService


@pytest.fixture(autouse=True)
def sanitizer_service(event_manager):
    return SanitizerService(event_manager)


@pytest.fixture(autouse=True)
def undo_manager(event_manager, element_factory):
    return UndoManager(event_manager, element_factory)


def test_remove_class_with_association(create, diagram, element_factory, event_manager):
    with Transaction(event_manager):
        c1 = create(ClassItem, UML.Class)
        c1.subject.name = "klassitem1"
        c2 = create(ClassItem, UML.Class)
        c2.subject.name = "klassitem2"

        a = create(AssociationItem)

        assert len(list(diagram.get_all_items())) == 3

        connect(a, a.head, c1)
        connect(a, a.tail, c2)

    assert a.subject
    assert element_factory.lselect(UML.Association)[0] is a.subject

    with Transaction(event_manager):
        c1.unlink()



from gi.repository import Gdk

from gaphor import UML
from gaphor.diagram.propertypages import PropertyPages
from gaphor.diagram.tests.fixtures import find
from gaphor.UML.actions.activitypropertypage import (
    ActivityPage,
    ActivityParameterNodeDirectionPropertyPage,
    ActivityParameterNodeNamePropertyPage,
    ActivityParameterNodeTypePropertyPage,
    activity_parameter_node_model,
    list_view_key_handler,
)
from gaphor.UML.propertypages import TypedElementPropertyPage


def activity_parameter_node(element_factory, name=None):
    node = element_factory.create(UML.ActivityParameterNode)
    node.parameter = element_factory.create(UML.Parameter)
    node.name = name
    return node


def test_activity_parameter_node_empty(element_factory, event_manager):
    activity = element_factory.create(UML.Activity)
    model = activity_parameter_node_model(activity, event_manager)

    empty = model.get_item(0)

    assert not empty.node


def test_activity_parameter_node_in_model(element_factory, event_manager):
    activity = element_factory.create(UML.Activity)
    activity.node = activity_parameter_node(element_factory, "name")
    model = activity_parameter_node_model(activity, event_manager)

    param = model.get_item(0)
    empty = model.get_item(1)

    assert param.node is activity.node[0]
    assert not empty.node


def test_activity_parameter_node_edit_existing_parameter(
    element_factory, event_manager
):
    activity = element_factory.create(UML.Activity)
    activity.node = activity_parameter_node(element_factory)
    parameter = activity.node[0].parameter
    model = activity_parameter_node_model(activity, event_manager)

    view = model.get_item(0)
    view.parameter = "in attr: str"

    assert parameter.direction == "in"
    assert parameter.name == "attr"
    assert parameter.typeValue == "str"


def test_activity_parameter_node_add_new_parameter(element_factory, event_manager):
    activity = element_factory.create(UML.Activity)
    model = activity_parameter_node_model(activity, event_manager)

    view = model.get_item(0)
    view.parameter = "in attr: str"
    parameter = activity.node[0].parameter

    assert view.node in activity.node
    assert parameter.direction == "in"
    assert parameter.name == "attr"
    assert parameter.typeValue == "str"


def test_update_model_when_new_parameter_added(element_factory, event_manager):
    subject = element_factory.create(UML.Activity)

    property_page = ActivityPage(subject, event_manager)
    property_page.construct()

    subject.node = activity_parameter_node(element_factory, "one")

    assert property_page.model.get_n_items() == 2  # one + empty row


def test_activity_parameter_node_name_editing(element_factory, event_manager):
    subject = element_factory.create(UML.ActivityParameterNode)
    subject.parameter = element_factory.create(UML.Parameter)
    property_page = ActivityParameterNodeNamePropertyPage(subject, event_manager)

    widget = property_page.construct()
    name = find(widget, "name-entry")
    name.set_text("A new name")

    assert subject.parameter.name == "A new name"


def test_update_model_when_new_parameter_added_via_list_model(
    element_factory, event_manager
):
    subject = element_factory.create(UML.Activity)

    property_page = ActivityPage(subject, event_manager)
    property_page.construct()

    new_view = property_page.model.get_item(0)
    new_view.parameter = "new"

    assert property_page.model.get_n_items() == 2  # one + empty row
    assert property_page.model.get_item(0).node is new_view.node
    assert property_page.model.get_item(1).node is None


def test_update_model_with_single_node_when_parameter_deleted(
    element_factory, event_manager
):
    subject = element_factory.create(UML.Activity)
    subject.node = node = activity_parameter_node(element_factory, "one")

    property_page = ActivityPage(subject, event_manager)
    property_page.construct()

    del subject.node[node]

    assert not subject.node
    assert property_page.model.get_n_items() == 1
    assert property_page.model.get_item(0).node is None


def test_update_model_with_multiple_nodes_when_parameter_deleted(
    element_factory, event_manager
):
    subject = element_factory.create(UML.Activity)
    subject.node = node = activity_parameter_node(element_factory, "one")
    subject.node = activity_parameter_node(element_factory, "two")

    property_page = ActivityPage(subject, event_manager)
    property_page.construct()

    del subject.node[node]

    assert property_page.model.get_n_items() == 2
    assert property_page.model.get_item(0).node in subject.node
    assert property_page.model.get_item(1).node is None


def test_activity_parameter_node_reorder_down(element_factory, event_manager):
    subject = element_factory.create(UML.Activity)
    subject.node = activity_parameter_node(element_factory, "one")
    subject.node = activity_parameter_node(element_factory, "two")
    node_one, node_two = subject.node

    property_page = ActivityPage(subject, event_manager)
    widget = property_page.construct()

    list_view = find(widget, "parameter-list")

    list_view_key_handler(ControllerStub(list_view), Gdk.KEY_plus, None, 0)

    assert subject.node[0] is node_two
    assert subject.node[1] is node_one


def test_activity_parameter_node_reorder_up(element_factory, event_manager):
    subject = element_factory.create(UML.Activity)
    subject.node = activity_parameter_node(element_factory, "one")
    subject.node = activity_parameter_node(element_factory, "two")
    node_one, node_two = subject.node

    property_page = ActivityPage(subject, event_manager)
    widget = property_page.construct()

    list_view = find(widget, "parameter-list")
    selection = list_view.get_model()
    selection.set_selected(1)

    list_view_key_handler(ControllerStub(list_view), Gdk.KEY_minus, None, 0)

    assert subject.node[0] is node_two
    assert subject.node[1] is node_one


def test_activity_parameter_node_reorder_first_node_up(element_factory, event_manager):
    subject = element_factory.create(UML.Activity)
    subject.node = activity_parameter_node(element_factory, "one")
    subject.node = activity_parameter_node(element_factory, "two")
    node_one, node_two = subject.node

    property_page = ActivityPage(subject, event_manager)
    widget = property_page.construct()

    list_view = find(widget, "parameter-list")

    list_view_key_handler(ControllerStub(list_view), Gdk.KEY_minus, None, 0)

    assert subject.node[0] is node_one
    assert subject.node[1] is node_two


def test_construct_activity_item_property_page(element_factory, event_manager):
    subject = element_factory.create(UML.Activity)

    property_page = ActivityPage(subject, event_manager)
    widget = property_page.construct()

    assert widget


def test_construct_activity_parameter_node_type_property_page(
    element_factory, event_manager
):
    subject = element_factory.create(UML.ActivityParameterNode)
    subject.parameter = element_factory.create(UML.Parameter)

    property_page = ActivityParameterNodeTypePropertyPage(subject, event_manager)
    widget = property_page.construct()

    assert widget


def test_construct_activity_parameter_node_direction_property_page(
    element_factory, event_manager
):
    subject = element_factory.create(UML.ActivityParameterNode)
    subject.parameter = element_factory.create(UML.Parameter)

    property_page = ActivityParameterNodeDirectionPropertyPage(subject, event_manager)
    widget = property_page.construct()

    assert widget


def test_construct_activity_parameter_node_direction_changed(
    element_factory, event_manager
):
    subject = element_factory.create(UML.ActivityParameterNode)
    subject.parameter = element_factory.create(UML.Parameter)

    property_page = ActivityParameterNodeDirectionPropertyPage(subject, event_manager)
    widget = property_page.construct()

    direction = find(widget, "parameter-direction")
    direction.set_selected(3)

    assert subject.parameter.direction == "return"


def test_one_typed_element_property_page_for_activity_parameter_node(element_factory):
    subject = element_factory.create(UML.ActivityParameterNode)
    subject.parameter = element_factory.create(UML.Parameter)

    property_pages = list(PropertyPages.find(subject))

    assert TypedElementPropertyPage not in property_pages
    assert ActivityParameterNodeTypePropertyPage in property_pages


class ControllerStub:
    def __init__(self, widget):
        self._widget = widget

    def get_widget(self):
        return self._widget



"""Test classes."""

from gaphor import UML
from gaphor.core.modeling import UpdateContext
from gaphor.core.modeling.diagram import Diagram
from gaphor.UML.classes.klass import ClassItem


def compartments(item):
    return item.shape.children[1:]


def context():
    return UpdateContext(style={})


def test_compartments(element_factory):
    """Test creation of classes and working of compartments."""
    diagram = element_factory.create(Diagram)
    klass = diagram.create(ClassItem, subject=element_factory.create(UML.Class))

    assert 2 == len(compartments(klass))
    assert 0 == len(compartments(klass)[0].child.children)
    assert 0 == len(compartments(klass)[1].child.children)

    diagram.update({klass})

    assert 50 == float(klass.min_height)
    assert 100 == float(klass.min_width)

    attr = element_factory.create(UML.Property)
    attr.name = 4 * "x"  # about 44 pixels
    klass.subject.ownedAttribute = attr

    diagram.update({klass})

    assert 1 == len(compartments(klass)[0].child.children)
    assert compartments(klass)[0].size(context()) >= (40, 15)

    oper = element_factory.create(UML.Operation)
    oper.name = 4 * "x"  # about 44 pixels
    klass.subject.ownedOperation = oper

    oper = element_factory.create(UML.Operation)
    oper.name = 6 * "x"  # about 66 pixels
    klass.subject.ownedOperation = oper

    diagram.update({klass})
    assert 2 == len(compartments(klass)[1].child.children)
    assert compartments(klass)[1].size(context()) > (60.0, 34.0)


def test_attribute_removal(element_factory):
    diagram = element_factory.create(Diagram)
    klass = diagram.create(ClassItem, subject=element_factory.create(UML.Class))
    diagram.update({klass})

    attr = element_factory.create(UML.Property)
    attr.name = "blah1"
    klass.subject.ownedAttribute = attr

    attr2 = element_factory.create(UML.Property)
    attr2.name = "blah2"
    klass.subject.ownedAttribute = attr2

    attr = element_factory.create(UML.Property)
    attr.name = "blah3"
    klass.subject.ownedAttribute = attr

    assert len(compartments(klass)[0].child.children) == 3

    attr2.unlink()

    assert len(compartments(klass)[0].child.children) == 2


def test_compartment_resizing(element_factory):
    diagram = element_factory.create(Diagram)
    klass = diagram.create(ClassItem, subject=element_factory.create(UML.Class))
    klass.subject.name = "Class1"

    diagram.update({klass})

    attr = element_factory.create(UML.Property)
    attr.name = "blah"
    klass.subject.ownedAttribute = attr

    oper = element_factory.create(UML.Operation)
    oper.name = "method"
    klass.subject.ownedOperation = oper

    assert klass.width == 100

    attr.name = "x" * 25

    diagram.update({klass})

    width = klass.width
    assert width >= 170.0



import pytest

import gaphor.UML.sanitizerservice
from gaphor import UML
from gaphor.application import Session
from gaphor.core import Transaction
from gaphor.diagram.tests.fixtures import connect
from gaphor.UML import diagramitems


@pytest.fixture
def session():
    session = Session()

    yield session
    session.shutdown()


@pytest.fixture
def event_manager(session):
    return session.get_service("event_manager")


@pytest.fixture
def element_factory(session):
    return session.get_service("element_factory")


@pytest.fixture
def properties(session):
    return session.get_service("properties")


@pytest.fixture
def diagram(event_manager, element_factory):
    with Transaction(event_manager):
        return element_factory.create(UML.Diagram)


@pytest.fixture
def classes_and_association(diagram, event_manager, element_factory):
    with Transaction(event_manager):
        c1 = diagram.create(
            diagramitems.ClassItem, subject=element_factory.create(UML.Class)
        )
        c2 = diagram.create(
            diagramitems.ClassItem, subject=element_factory.create(UML.Class)
        )

        a = diagram.create(diagramitems.AssociationItem)
        connect(a, a.handles()[0], c1)
        connect(a, a.handles()[1], c2)

    return c1.subject, c2.subject, a.subject


class MockSettings:
    def __init__(self, value):
        self._value = value

    @property
    def remove_unused_elements(self):
        return self._value


@pytest.mark.parametrize(
    ["remove_unused_elements", "comparator"],
    [
        [True, lambda a: not a],
        [False, lambda a: a],
    ],
)
def test_delete_diagram(
    classes_and_association,
    diagram,
    event_manager,
    element_factory,
    monkeypatch,
    remove_unused_elements,
    comparator,
):
    monkeypatch.setattr(
        gaphor.UML.sanitizerservice, "settings", MockSettings(remove_unused_elements)
    )

    c1, c2, a = classes_and_association

    with Transaction(event_manager):
        diagram.unlink()

    assert comparator(c1 in element_factory)
    assert comparator(c2 in element_factory)
    assert comparator(a in element_factory)



import pytest

from gaphor.core.format import format
from gaphor.SysML import sysml

TYPES = [sysml.Allocate, sysml.Refine, sysml.Trace]


@pytest.mark.parametrize("type", TYPES)
def test_format_simple_property_path(type):
    rel = type()

    assert format(rel) == "sourceContext: "


@pytest.mark.parametrize("type", TYPES)
def test_format_connected_property_path(type):
    block = sysml.Block()
    block.name = "MyBlock"

    rel = type()
    rel.sourceContext = block

    assert format(rel) == "sourceContext: MyBlock"



from gaphor import UML
from gaphor.C4Model.c4model import Container
from gaphor.C4Model.diagramitems import C4ContainerItem
from gaphor.diagram.group import group


def container(diagram, element_factory):
    subject = element_factory.create(Container)
    return diagram.create(C4ContainerItem, subject=subject)


def test_group_container_can_contain(diagram, element_factory):
    parent = container(diagram, element_factory)
    child = container(diagram, element_factory)

    assert group(parent.subject, child.subject)


def test_group_container_can_group(diagram, element_factory):
    parent = container(diagram, element_factory)
    child = container(diagram, element_factory)

    group(parent.subject, child.subject)

    assert child.subject.namespace is parent.subject


def test_group_container_group_from_package_to_container(diagram, element_factory):
    package = element_factory.create(UML.Package)
    parent = container(diagram, element_factory)
    child = container(diagram, element_factory)

    child.subject.nestingPackage = package
    group(parent.subject, child.subject)

    assert child.subject.namespace is parent.subject


def test_group_container_ungroup(diagram, element_factory):
    package = element_factory.create(UML.Package)
    diagram.element = package
    parent = container(diagram, element_factory)
    child = container(diagram, element_factory)
    child.subject.nestingPackage = parent.subject

    did_group = group(package, child.subject)

    assert did_group
    assert child.subject.namespace is package



from gi.repository import GLib


def iterate_until(condition, timeout=5):
    sentinel = False

    def check_condition():
        nonlocal sentinel
        if condition():
            sentinel = True
        return GLib.SOURCE_REMOVE if sentinel else GLib.SOURCE_CONTINUE

    def do_timeout():
        nonlocal sentinel
        sentinel = True
        return GLib.SOURCE_REMOVE

    GLib.idle_add(check_condition, priority=GLib.PRIORITY_LOW)
    timeout_id = GLib.timeout_add(interval=timeout * 1_000, function=do_timeout)

    ctx = GLib.main_context_default()
    while ctx.pending():
        if sentinel:
            break
        ctx.iteration(False)

    if not sentinel:
        GLib.source_remove(timeout_id)



from gaphor import UML
from gaphor.diagram.tests.fixtures import find
from gaphor.UML.interactions.interactionspropertypages import (
    LifelinePropertyPage,
    MessagePropertyPage,
)


def test_lifeline_property_page(diagram, element_factory, event_manager):
    subject = element_factory.create(UML.Lifeline)

    type = element_factory.create(UML.Interface)
    type.name = "Bar"
    property_page = LifelinePropertyPage(subject, event_manager)

    widget = property_page.construct()
    dropdown = find(widget, "element-type")
    bar_index = next(
        n for n, lv in enumerate(dropdown.get_model()) if lv.value == type.id
    )
    dropdown.set_selected(bar_index)

    assert dropdown.get_selected_item().label == "Bar"
    assert subject.represents is type


def test_message_property_page(diagram, element_factory, event_manager):
    item = diagram.create(
        UML.interactions.MessageItem, subject=element_factory.create(UML.Message)
    )
    property_page = MessagePropertyPage(item, event_manager)

    widget = property_page.construct()
    message_combo = find(widget, "message-combo")
    message_combo.set_selected(2)

    assert item.subject.messageSort == "asynchSignal"



from pathlib import Path
from xml.etree import ElementTree as etree

import pytest
from markdown_it import MarkdownIt

# The release description field in a release supports a small subset of HTML:
# * p, ol, ul, li, em, code (no nested lists).
# * Only <p> and <li> may contain a `xml:lang` attribute.
#
# https://www.freedesktop.org/software/appstream/docs/chap-Metadata.html#tag-description


def versions(changelog_md: str):
    md = MarkdownIt("default", {"breaks": False, "html": True})
    html = md.render(changelog_md)
    changelog = etree.fromstring(f"<root>{html}</root>")

    version = "INVALID"
    news: dict[str, etree.Element] = {}

    for node in changelog:
        if node.tag == "h2" and node.text:
            version = node.text
            news[version] = etree.Element("description")
        else:
            news[version].append(node)

    return news


@pytest.mark.parametrize(
    "version,description",
    list(versions(Path("CHANGELOG.md").read_text("utf-8")).items()),
)
def test_changelog_is_appdata_compliant(version, description):
    whitelist_top = ["p", "ol", "ul", "em", "code"]
    whitelist_nested = ["li", "em", "code"]

    def _test_nested(elem):
        for e in elem:
            assert e.tag in whitelist_nested
            _test_nested(e)

    for e in description:
        assert e.tag in whitelist_top
        _test_nested(e)



from gaphor.ui.statuswindow import StatusWindow


def test_status_window():
    status_window = StatusWindow("title", "message")

    assert status_window.progress_bar
    assert status_window.window



import gaphas
import pytest

from gaphor.core import event_handler
from gaphor.core.modeling import Base, Diagram, ElementDeleted, Presentation
from gaphor.core.modeling.properties import association


class Example(gaphas.Element, Presentation):
    def __init__(self, diagram, id):
        super().__init__(connections=diagram.connections, diagram=diagram, id=id)
        self._test_unlinked = False

    @property
    def test_unlinked(self):
        return self._test_unlinked

    def unlink(self):
        self._test_unlinked = True
        super().unlink()


class ExampleLine(gaphas.Line, Presentation):
    def __init__(self, diagram, id):
        super().__init__(connections=diagram.connections, diagram=diagram, id=id)
        self._test_unlinked = False

    @property
    def test_unlinked(self):
        return self._test_unlinked

    def unlink(self):
        self._test_unlinked = True
        super().unlink()


class ModelElement(Base):
    diagram = association("association", Diagram, composite=True)


@pytest.fixture
def diagram(element_factory):
    return element_factory.create(Diagram)


def test_diagram_can_be_used_as_gtkview_model():
    diagram = Diagram("id", None)

    assert isinstance(diagram, gaphas.model.Model)


def test_canvas_is_saved():
    diagram = Diagram("id", None)
    saved_keys = []
    diagram.save(lambda name, val: saved_keys.append(name))

    assert "canvas" not in saved_keys


def test_canvas_item_is_created(element_factory):
    diagram = element_factory.create(Diagram)
    example = diagram.create(Example)

    assert example in diagram.get_all_items()
    assert example.diagram is diagram


def test_canvas_is_unlinked(element_factory):
    diagram = element_factory.create(Diagram)
    example = diagram.create(Example)

    diagram.unlink()

    assert example.test_unlinked


def test_can_only_add_diagram_items(element_factory):
    diagram = element_factory.create(Diagram)

    with pytest.raises(TypeError):
        diagram.create(Diagram)


class ViewMock:
    def __init__(self):
        self.removed_items = set()

    def request_update(self, items, removed_items) -> None:
        self.removed_items.update(removed_items)


def test_remove_presentation_triggers_view(element_factory):
    diagram = element_factory.create(Diagram)
    view = ViewMock()
    diagram.register_view(view)

    example = diagram.create(Example)

    example.unlink()

    assert example.diagram is None
    assert example not in diagram.ownedPresentation
    assert example in view.removed_items


def test_order_presentations_lines_are_last(diagram):
    example_line = diagram.create(ExampleLine)
    example = diagram.create(Example)

    assert list(diagram.get_all_items()) == [example, example_line]


def test_order_presentations_line_is_grouped(diagram):
    example_line = diagram.create(ExampleLine)
    example_1 = diagram.create(Example)
    example_2 = diagram.create(Example)

    example_line.parent = example_1

    assert list(diagram.get_all_items()) == [example_1, example_2, example_line]


def test_order_grouped_presentations(diagram):
    example_1 = diagram.create(Example)
    example_2 = diagram.create(Example)

    example_1.parent = example_2

    assert list(diagram.get_all_items()) == [example_2, example_1]


def test_unlink_presentations_before_diagram(diagram, event_manager):
    events = []

    @event_handler(ElementDeleted)
    def listener(event):
        events.append(event)

    event_manager.subscribe(listener)

    diagram.create(Example)
    diagram.create(Example)

    diagram.unlink()

    assert events[-1].element is diagram


def test_unlink_order_with_parent_element_in_diagram(
    diagram, event_manager, element_factory, sanitizer_service
):
    events = []

    @event_handler(ElementDeleted)
    def listener(event):
        events.append(event)

    event_manager.subscribe(listener)

    element = element_factory.create(ModelElement)
    example = diagram.create(Example, subject=element)
    element.diagram = diagram

    example.unlink()

    assert events[-1].element is not example



import logging
import sys
from pathlib import Path

import pytest

import gaphor.ui
from gaphor.main import main

APP_NAME = sys.argv[0]


@pytest.fixture(autouse=True)
def mock_gaphor_ui_run(monkeypatch):
    _argv = []

    def run(argv, recover=False):
        _argv[:] = argv
        return 0

    monkeypatch.setattr(gaphor.ui, "run", run)
    return _argv


def test_run_main():
    exit_code = main([APP_NAME])

    assert exit_code == 0


def test_version(capsys):
    with pytest.raises(SystemExit):
        main([APP_NAME, "-V"])

    assert "Gaphor" in capsys.readouterr().out


def test_debug_logging():
    main([APP_NAME, "-v"])

    assert logging.getLogger("gaphor").getEffectiveLevel() == logging.DEBUG


def test_quiet_logging():
    main([APP_NAME, "-q"])

    assert logging.getLogger("root").getEffectiveLevel() == logging.WARNING


def test_gapplication_service(mock_gaphor_ui_run):
    main([APP_NAME, "--gapplication-service"])

    assert mock_gaphor_ui_run == [APP_NAME, "--gapplication-service"]


def test_run_script(capsys):
    run_script = Path(__file__).parent / "run_script.py"

    main([APP_NAME, "exec", str(run_script)])

    assert "Running a test script for Gaphor" in capsys.readouterr().out
